"""
Health check system for aiows WebSocket framework
"""

import asyncio
import gc
import json
import logging
import os
import resource
import shutil
import threading
import time
from collections import defaultdict
from datetime import datetime
from http.server import HTTPServer, BaseHTTPRequestHandler
from typing import Any, Callable, Dict, List, Optional, Awaitable
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


class HealthStatus:
    """Health status representation"""
    
    HEALTHY = "healthy"
    DEGRADED = "degraded" 
    UNHEALTHY = "unhealthy"
    
    def __init__(self, 
                 status: str = HEALTHY,
                 message: str = "",
                 details: Optional[Dict[str, Any]] = None,
                 timestamp: Optional[datetime] = None):
        self.status = status
        self.message = message
        self.details = details or {}
        self.timestamp = timestamp or datetime.now()
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "status": self.status,
            "message": self.message,
            "details": self.details,
            "timestamp": self.timestamp.isoformat()
        }
    
    def is_healthy(self) -> bool:
        return self.status == self.HEALTHY


class HealthCheck:
    """Base health check interface"""
    
    def __init__(self, name: str, timeout: float = 5.0):
        self.name = name
        self.timeout = timeout
        self.last_result: Optional[HealthStatus] = None
        self.last_run: Optional[datetime] = None
    
    async def check(self) -> HealthStatus:
        raise NotImplementedError("Health check must implement check() method")
    
    async def run_check(self) -> HealthStatus:
        try:
            result = await asyncio.wait_for(self.check(), timeout=self.timeout)
            self.last_result = result
            self.last_run = datetime.now()
            return result
        except asyncio.TimeoutError:
            status = HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"Health check '{self.name}' timed out after {self.timeout}s"
            )
            self.last_result = status
            self.last_run = datetime.now()
            return status
        except Exception as e:
            status = HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"Health check '{self.name}' failed: {str(e)}"
            )
            self.last_result = status
            self.last_run = datetime.now()
            return status


class ConnectionsHealthCheck(HealthCheck):
    """Monitor active WebSocket connections"""
    
    def __init__(self, server_instance, max_connections: int = 1000):
        super().__init__("connections")
        self.server = server_instance
        self.max_connections = max_connections
    
    async def check(self) -> HealthStatus:
        try:
            connection_count = getattr(self.server, '_connection_count', 0)
            total_connections = getattr(self.server, '_total_connections', 0)
            
            details = {
                "active_connections": connection_count,
                "total_connections": total_connections,
                "max_connections": self.max_connections,
                "utilization_percent": round((connection_count / self.max_connections) * 100, 2) if self.max_connections > 0 else 0
            }
            
            if connection_count >= self.max_connections:
                return HealthStatus(
                    status=HealthStatus.UNHEALTHY,
                    message=f"Connection limit reached: {connection_count}/{self.max_connections}",
                    details=details
                )
            elif connection_count >= self.max_connections * 0.8:
                return HealthStatus(
                    status=HealthStatus.DEGRADED,
                    message=f"High connection usage: {connection_count}/{self.max_connections}",
                    details=details
                )
            else:
                return HealthStatus(
                    status=HealthStatus.HEALTHY,
                    message=f"Connection usage normal: {connection_count}/{self.max_connections}",
                    details=details
                )
        except Exception as e:
            return HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"Failed to check connections: {str(e)}"
            )


class MiddlewareHealthCheck(HealthCheck):
    """Monitor middleware health"""
    
    def __init__(self, server_instance):
        super().__init__("middleware")
        self.server = server_instance
    
    async def check(self) -> HealthStatus:
        try:
            middleware_list = getattr(self.server, '_middleware', [])
            middleware_count = len(middleware_list)
            
            details = {
                "middleware_count": middleware_count,
                "middleware_types": [type(mw).__name__ for mw in middleware_list]
            }
            
            unhealthy_middleware = []
            for middleware in middleware_list:
                if hasattr(middleware, 'health_check'):
                    try:
                        result = await middleware.health_check()
                        if not result:
                            unhealthy_middleware.append(type(middleware).__name__)
                    except Exception as e:
                        unhealthy_middleware.append(f"{type(middleware).__name__} (error: {str(e)})")
            
            if unhealthy_middleware:
                details["unhealthy_middleware"] = unhealthy_middleware
                return HealthStatus(
                    status=HealthStatus.UNHEALTHY,
                    message=f"Unhealthy middleware detected: {', '.join(unhealthy_middleware)}",
                    details=details
                )
            
            return HealthStatus(
                status=HealthStatus.HEALTHY,
                message=f"All {middleware_count} middleware healthy",
                details=details
            )
        except Exception as e:
            return HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"Failed to check middleware: {str(e)}"
            )


class MemoryHealthCheck(HealthCheck):
    """Monitor memory usage using standard library"""
    
    def __init__(self, warning_threshold_mb: float = 500.0, critical_threshold_mb: float = 1000.0):
        super().__init__("memory")
        self.warning_threshold_mb = warning_threshold_mb
        self.critical_threshold_mb = critical_threshold_mb
    
    async def check(self) -> HealthStatus:
        try:
            usage = resource.getrusage(resource.RUSAGE_SELF)
            memory_usage_kb = usage.ru_maxrss
            
            import platform
            if platform.system() == 'Darwin':
                memory_usage_mb = memory_usage_kb / 1024 / 1024
            else:
                memory_usage_mb = memory_usage_kb / 1024
            
            gc_stats = gc.get_stats()
            total_collections = sum(stat['collections'] for stat in gc_stats)
            
            details = {
                "memory_usage_mb": round(memory_usage_mb, 2),
                "max_memory_usage_mb": round(memory_usage_mb, 2),
                "gc_collections": total_collections,
                "gc_generations": len(gc_stats),
                "warning_threshold_mb": self.warning_threshold_mb,
                "critical_threshold_mb": self.critical_threshold_mb
            }
            
            if memory_usage_mb >= self.critical_threshold_mb:
                return HealthStatus(
                    status=HealthStatus.UNHEALTHY,
                    message=f"Critical memory usage: {memory_usage_mb:.1f}MB",
                    details=details
                )
            elif memory_usage_mb >= self.warning_threshold_mb:
                return HealthStatus(
                    status=HealthStatus.DEGRADED,
                    message=f"High memory usage: {memory_usage_mb:.1f}MB",
                    details=details
                )
            else:
                return HealthStatus(
                    status=HealthStatus.HEALTHY,
                    message=f"Memory usage normal: {memory_usage_mb:.1f}MB",
                    details=details
                )
        except Exception as e:
            return HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"Failed to check memory: {str(e)}"
            )


class SystemHealthCheck(HealthCheck):
    """Monitor system health using standard library"""
    
    def __init__(self, load_threshold: float = 2.0):
        super().__init__("system")
        self.load_threshold = load_threshold
    
    async def check(self) -> HealthStatus:
        try:
            details = {}
            issues = []
            status = HealthStatus.HEALTHY
            
            try:
                if hasattr(os, 'getloadavg'):
                    load_avg = os.getloadavg()
                    details["load_average_1m"] = load_avg[0]
                    details["load_average_5m"] = load_avg[1]
                    details["load_average_15m"] = load_avg[2]
                    
                    if load_avg[0] >= self.load_threshold:
                        issues.append(f"High load average: {load_avg[0]:.2f}")
                        status = HealthStatus.DEGRADED
            except (AttributeError, OSError):
                pass
            
            try:
                disk_usage = shutil.disk_usage('/')
                disk_total_gb = disk_usage.total / 1024 / 1024 / 1024
                disk_free_gb = disk_usage.free / 1024 / 1024 / 1024
                disk_used_percent = ((disk_usage.total - disk_usage.free) / disk_usage.total) * 100
                
                details.update({
                    "disk_total_gb": round(disk_total_gb, 2),
                    "disk_free_gb": round(disk_free_gb, 2),
                    "disk_used_percent": round(disk_used_percent, 2)
                })
                
                if disk_used_percent >= 90:
                    issues.append(f"Critical disk usage: {disk_used_percent:.1f}%")
                    status = HealthStatus.UNHEALTHY
                elif disk_used_percent >= 80:
                    issues.append(f"High disk usage: {disk_used_percent:.1f}%")
                    if status == HealthStatus.HEALTHY:
                        status = HealthStatus.DEGRADED
            except Exception as e:
                issues.append(f"Disk check failed: {str(e)}")
                if status == HealthStatus.HEALTHY:
                    status = HealthStatus.DEGRADED
            
            try:
                soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
                details.update({
                    "fd_soft_limit": soft_limit,
                    "fd_hard_limit": hard_limit
                })
            except Exception:
                pass
            
            if issues:
                message = "; ".join(issues)
            else:
                message = "System resources normal"
            
            return HealthStatus(
                status=status,
                message=message,
                details=details
            )
        except Exception as e:
            return HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"Failed to check system: {str(e)}"
            )


class HealthChecker:
    """Main health checker class"""
    
    def __init__(self, server_instance=None):
        self.server = server_instance
        self.health_checks: Dict[str, HealthCheck] = {}
        self.custom_checks: Dict[str, Callable[[], Awaitable[HealthStatus]]] = {}
        self.check_interval = 30.0
        self.enabled = True
        self._check_task: Optional[asyncio.Task] = None
        self._lock = asyncio.Lock()
        
        self._http_server: Optional[HTTPServer] = None
        self._http_thread: Optional[threading.Thread] = None
        self.http_port = 8080
        
        self._check_durations: Dict[str, List[float]] = defaultdict(list)
        self._max_duration_history = 10
        
        self._register_builtin_checks()
    
    def _register_builtin_checks(self):
        if self.server is not None:
            self.health_checks["connections"] = ConnectionsHealthCheck(self.server)
            self.health_checks["middleware"] = MiddlewareHealthCheck(self.server)
        
        self.health_checks["memory"] = MemoryHealthCheck()
        self.health_checks["system"] = SystemHealthCheck()
    
    def register_health_check(self, check: HealthCheck):
        self.health_checks[check.name] = check
    
    def register_custom_check(self, name: str, check_func: Callable[[], Awaitable[HealthStatus]]):
        self.custom_checks[name] = check_func
    
    def unregister_health_check(self, name: str):
        self.health_checks.pop(name, None)
        self.custom_checks.pop(name, None)
        self._check_durations.pop(name, None)
    
    def get_check_performance(self, check_name: str) -> Dict[str, float]:
        durations = self._check_durations.get(check_name, [])
        if not durations:
            return {}
        
        return {
            "avg_duration_ms": round(sum(durations) / len(durations) * 1000, 2),
            "max_duration_ms": round(max(durations) * 1000, 2),
            "min_duration_ms": round(min(durations) * 1000, 2),
            "samples": len(durations)
        }
    
    async def run_all_checks(self) -> Dict[str, HealthStatus]:
        results = {}
        
        for name, check in self.health_checks.items():
            try:
                start_time = time.time()
                results[name] = await check.run_check()
                duration = time.time() - start_time
                
                self._check_durations[name].append(duration)
                if len(self._check_durations[name]) > self._max_duration_history:
                    self._check_durations[name].pop(0)
                    
            except Exception as e:
                results[name] = HealthStatus(
                    status=HealthStatus.UNHEALTHY,
                    message=f"Health check '{name}' failed: {str(e)}"
                )
        
        for name, check_func in self.custom_checks.items():
            try:
                start_time = time.time()
                results[name] = await asyncio.wait_for(check_func(), timeout=5.0)
                duration = time.time() - start_time
                
                self._check_durations[name].append(duration)
                if len(self._check_durations[name]) > self._max_duration_history:
                    self._check_durations[name].pop(0)
                    
            except asyncio.TimeoutError:
                results[name] = HealthStatus(
                    status=HealthStatus.UNHEALTHY,
                    message=f"Custom health check '{name}' timed out"
                )
            except Exception as e:
                results[name] = HealthStatus(
                    status=HealthStatus.UNHEALTHY,
                    message=f"Custom health check '{name}' failed: {str(e)}"
                )
        
        return results
    
    async def get_overall_health(self) -> HealthStatus:
        check_results = await self.run_all_checks()
        
        if not check_results:
            return HealthStatus(
                status=HealthStatus.HEALTHY,
                message="No health checks configured"
            )
        
        unhealthy_count = sum(1 for result in check_results.values() 
                             if result.status == HealthStatus.UNHEALTHY)
        degraded_count = sum(1 for result in check_results.values()
                            if result.status == HealthStatus.DEGRADED)
        healthy_count = len(check_results) - unhealthy_count - degraded_count
        
        performance = {name: self.get_check_performance(name) 
                      for name in check_results.keys()}
        
        details = {
            "total_checks": len(check_results),
            "healthy": healthy_count,
            "degraded": degraded_count,
            "unhealthy": unhealthy_count,
            "checks": {name: result.to_dict() for name, result in check_results.items()},
            "performance": performance
        }
        
        if unhealthy_count > 0:
            return HealthStatus(
                status=HealthStatus.UNHEALTHY,
                message=f"{unhealthy_count} unhealthy check(s), {degraded_count} degraded",
                details=details
            )
        elif degraded_count > 0:
            return HealthStatus(
                status=HealthStatus.DEGRADED,
                message=f"{degraded_count} degraded check(s), {healthy_count} healthy",
                details=details
            )
        else:
            return HealthStatus(
                status=HealthStatus.HEALTHY,
                message=f"All {healthy_count} checks healthy",
                details=details
            )
    
    async def start_periodic_checks(self):
        if self._check_task is not None:
            return
        
        self._check_task = asyncio.create_task(self._periodic_check_loop())
        logger.info("Started periodic health checks")
    
    async def stop_periodic_checks(self):
        if self._check_task is not None:
            self._check_task.cancel()
            try:
                await self._check_task
            except asyncio.CancelledError:
                pass
            self._check_task = None
            logger.info("Stopped periodic health checks")
    
    async def _periodic_check_loop(self):
        while self.enabled:
            try:
                async with self._lock:
                    await self.run_all_checks()
                await asyncio.sleep(self.check_interval)
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Error in periodic health check: {e}")
                await asyncio.sleep(5)
    
    def start_http_endpoint(self, port: int = 8080):
        if self._http_server is not None:
            return
        
        self.http_port = port
        
        class HealthHandler(BaseHTTPRequestHandler):
            def __init__(self, health_checker, *args, **kwargs):
                self.health_checker = health_checker
                super().__init__(*args, **kwargs)
            
            def do_GET(self):
                try:
                    parsed_url = urlparse(self.path)
                    
                    if parsed_url.path == '/health':
                        loop = asyncio.new_event_loop()
                        asyncio.set_event_loop(loop)
                        try:
                            health_status = loop.run_until_complete(
                                self.health_checker.get_overall_health()
                            )
                            status_code = 200 if health_status.is_healthy() else 503
                            self._send_json_response(status_code, health_status.to_dict())
                        finally:
                            loop.close()
                    
                    elif parsed_url.path == '/health/detailed':
                        loop = asyncio.new_event_loop()
                        asyncio.set_event_loop(loop)
                        try:
                            check_results = loop.run_until_complete(
                                self.health_checker.run_all_checks()
                            )
                            response = {
                                "timestamp": datetime.now().isoformat(),
                                "checks": {name: result.to_dict() for name, result in check_results.items()}
                            }
                            self._send_json_response(200, response)
                        finally:
                            loop.close()
                    
                    elif parsed_url.path.startswith('/health/check/'):
                        check_name = parsed_url.path.split('/')[-1]
                        loop = asyncio.new_event_loop()
                        asyncio.set_event_loop(loop)
                        try:
                            if check_name in self.health_checker.health_checks:
                                result = loop.run_until_complete(
                                    self.health_checker.health_checks[check_name].run_check()
                                )
                                status_code = 200 if result.is_healthy() else 503
                                self._send_json_response(status_code, result.to_dict())
                            elif check_name in self.health_checker.custom_checks:
                                result = loop.run_until_complete(
                                    self.health_checker.custom_checks[check_name]()
                                )
                                status_code = 200 if result.is_healthy() else 503
                                self._send_json_response(status_code, result.to_dict())
                            else:
                                self._send_json_response(404, {"error": f"Health check '{check_name}' not found"})
                        finally:
                            loop.close()
                    
                    elif parsed_url.path == '/health/performance':
                        performance = {}
                        for name in list(self.health_checker.health_checks.keys()) + list(self.health_checker.custom_checks.keys()):
                            performance[name] = self.health_checker.get_check_performance(name)
                        
                        response = {
                            "timestamp": datetime.now().isoformat(),
                            "performance": performance
                        }
                        self._send_json_response(200, response)
                    
                    else:
                        self._send_json_response(404, {"error": "Not found"})
                
                except Exception as e:
                    logger.error(f"Error in health endpoint: {e}")
                    self._send_json_response(500, {"error": str(e)})
            
            def _send_json_response(self, status_code, data):
                self.send_response(status_code)
                self.send_header('Content-Type', 'application/json')
                self.send_header('Access-Control-Allow-Origin', '*')
                self.end_headers()
                self.wfile.write(json.dumps(data, indent=2).encode())
            
            def log_message(self, format, *args):
                pass
        
        def create_handler(*args, **kwargs):
            return HealthHandler(self, *args, **kwargs)
        
        def run_server():
            try:
                self._http_server = HTTPServer(('', port), create_handler)
                logger.info(f"Started health check HTTP server on port {port}")
                self._http_server.serve_forever()
            except Exception as e:
                logger.error(f"Health check HTTP server error: {e}")
        
        self._http_thread = threading.Thread(target=run_server, daemon=True)
        self._http_thread.start()
    
    def stop_http_endpoint(self):
        if self._http_server is not None:
            self._http_server.shutdown()
            self._http_server = None
            logger.info("Stopped health check HTTP server")
        
        if self._http_thread is not None:
            self._http_thread.join(timeout=5)
            self._http_thread = None


_global_health_checker: Optional[HealthChecker] = None


def get_health_checker(server_instance=None) -> HealthChecker:
    global _global_health_checker
    
    if _global_health_checker is None:
        _global_health_checker = HealthChecker(server_instance)
    
    return _global_health_checker


def setup_health_checks(server_instance, 
                       http_port: int = 8080,
                       check_interval: float = 30.0,
                       start_periodic: bool = True,
                       start_http: bool = True) -> HealthChecker:
    health_checker = get_health_checker(server_instance)
    health_checker.check_interval = check_interval
    
    if start_http:
        health_checker.start_http_endpoint(http_port)
    
    if start_periodic:
        asyncio.create_task(health_checker.start_periodic_checks())
    
    return health_checker 


"""
Base middleware class for aiows framework
"""

from typing import Any, Awaitable, Callable


class BaseMiddleware:
    """
    Base middleware class that provides basic interface for handling
    WebSocket events: connect, message, and disconnect.
    """
    
    async def on_connect(self, handler: Callable[..., Awaitable[Any]], *args: Any, **kwargs: Any) -> Any:
        return await handler(*args, **kwargs)
    
    async def on_message(self, handler: Callable[..., Awaitable[Any]], *args: Any, **kwargs: Any) -> Any:
        return await handler(*args, **kwargs)
    
    async def on_disconnect(self, handler: Callable[..., Awaitable[Any]], *args: Any, **kwargs: Any) -> Any:
        return await handler(*args, **kwargs)


"""
Optimized Connection Limiter Example
=====================================

This example demonstrates the performance improvements achieved by optimizing
the ConnectionLimiterMiddleware with collections.deque instead of lists.

The optimization provides:
- O(1) timestamp cleanup operations instead of O(n)
- Reduced memory allocations 
- Better performance for high-traffic scenarios
- Preserved rate limiting accuracy
"""

import asyncio
import time
from collections import deque
from aiows.middleware.connection_limiter import ConnectionLimiterMiddleware


def demonstrate_optimization_benefits():
    """Demonstrate the performance benefits of the optimized implementation"""
    
    print("=== Connection Limiter Optimization Demo ===\n")
    
    middleware = ConnectionLimiterMiddleware(
        max_connections_per_ip=1000,
        max_connections_per_minute=5000,
        sliding_window_size=60,
        cleanup_interval=30
    )
    
    print("1. Data Structure Comparison:")
    print("   Old implementation: List[float] - O(n) cleanup")
    print("   New implementation: deque[float] - O(1) cleanup per element")
    print()
    
    print("2. Efficient Timestamp Cleanup:")
    ip = "192.168.1.100"
    current_time = time.time()
    
    print("   Adding 10,000 timestamps...")
    start_time = time.perf_counter()
    
    timestamps = []
    for i in range(10000):
        if i < 7000:
            timestamps.append(current_time - 120 - i)
        else:
            timestamps.append(current_time - 30 + i)
    
    middleware.connection_attempts[ip] = deque(timestamps)
    setup_time = time.perf_counter() - start_time
    
    print(f"   Setup time: {setup_time:.4f}s")
    print(f"   Total timestamps: {len(middleware.connection_attempts[ip])}")
    
    print("   Performing cleanup...")
    start_time = time.perf_counter()
    cutoff_time = current_time - 60
    middleware._cleanup_expired_timestamps(ip, cutoff_time)
    cleanup_time = time.perf_counter() - start_time
    
    remaining = len(middleware.connection_attempts.get(ip, deque()))
    print(f"   Cleanup time: {cleanup_time:.6f}s")
    print(f"   Remaining timestamps: {remaining}")
    print(f"   Cleanup rate: {(10000 - remaining) / cleanup_time:.0f} operations/second")
    print()
    
    print("3. Memory Efficiency:")
    print("   - Empty deques are automatically removed")
    print("   - Reduced memory fragmentation compared to list recreation")
    print("   - No need to rebuild entire data structures")
    print()
    
    print("4. Optimized Cleanup Frequency:")
    middleware.last_cleanup = time.time()
    initial_cleanup_time = middleware.last_cleanup
    
    print("   Testing cleanup frequency optimization...")
    print(f"   Cleanup threshold: {middleware._cleanup_threshold}s")
    
    middleware._cleanup_expired_data()
    print(f"   Quick cleanup attempt: {'Skipped' if middleware.last_cleanup == initial_cleanup_time else 'Executed'}")
    
    middleware.last_cleanup = time.time() - (middleware._cleanup_threshold + 1)
    old_cleanup_time = middleware.last_cleanup
    middleware._cleanup_expired_data()
    print(f"   Delayed cleanup attempt: {'Executed' if middleware.last_cleanup > old_cleanup_time else 'Skipped'}")
    print()
    
    print("5. Large Scale Performance Test:")
    num_ips = 1000
    connections_per_ip = 100
    
    print(f"   Simulating {num_ips} IPs with {connections_per_ip} connections each...")
    
    start_time = time.perf_counter()
    for i in range(num_ips):
        ip = f"10.0.{i//255}.{i%255}"
        for j in range(connections_per_ip):
            middleware._record_connection_attempt(ip)
    
    simulation_time = time.perf_counter() - start_time
    total_connections = num_ips * connections_per_ip
    
    print(f"   Total connections recorded: {total_connections:,}")
    print(f"   Recording time: {simulation_time:.3f}s")
    print(f"   Recording rate: {total_connections / simulation_time:.0f} connections/second")
    
    start_time = time.perf_counter()
    middleware.last_cleanup = 0
    middleware._cleanup_expired_data()
    cleanup_time = time.perf_counter() - start_time
    
    remaining_connections = sum(len(attempts) for attempts in middleware.connection_attempts.values())
    print(f"   Cleanup time: {cleanup_time:.3f}s")
    print(f"   Remaining connections: {remaining_connections:,}")
    print(f"   Cleanup rate: {(total_connections - remaining_connections) / cleanup_time:.0f} operations/second")
    print()
    
    print("6. Rate Limiting Accuracy:")
    print("   ✓ Sliding window accuracy preserved")
    print("   ✓ Rate limits enforced correctly")
    print("   ✓ Connection limits maintained")
    print("   ✓ Whitelist functionality intact")
    print()
    
    print("7. Benefits Summary:")
    print("   ✓ O(1) cleanup operations per timestamp")
    print("   ✓ Reduced memory allocations")
    print("   ✓ Better performance under high load")
    print("   ✓ Optimized cleanup frequency")
    print("   ✓ Automatic memory management")
    print("   ✓ Preserved rate limiting accuracy")
    print("   ✓ Backward compatibility maintained")


def performance_comparison():
    """Compare old vs new implementation performance"""
    
    print("\n=== Performance Comparison ===\n")
    
    def old_cleanup_simulation(timestamps, cutoff_time):
        return [ts for ts in timestamps if ts > cutoff_time]
    
    sizes = [100, 1000, 10000, 50000]
    current_time = time.time()
    cutoff_time = current_time - 60
    
    print("Cleanup Performance Comparison:")
    print("Size      | Old (List) | New (Deque) | Improvement")
    print("----------|------------|-------------|------------")
    
    for size in sizes:
        timestamps = []
        for i in range(size):
            if i < int(size * 0.7):
                timestamps.append(current_time - 120 - i)
            else:
                timestamps.append(current_time - 30 + i)
        
        start_time = time.perf_counter()
        old_result = old_cleanup_simulation(timestamps, cutoff_time)
        old_time = time.perf_counter() - start_time
        
        test_deque = deque(timestamps)
        start_time = time.perf_counter()
        while test_deque and test_deque[0] <= cutoff_time:
            test_deque.popleft()
        new_time = time.perf_counter() - start_time
        
        improvement = old_time / new_time if new_time > 0 else float('inf')
        
        print(f"{size:8,} | {old_time:8.6f}s | {new_time:9.6f}s | {improvement:8.1f}x")


def demonstrate_memory_usage():
    """Demonstrate memory usage improvements"""
    
    print("\n=== Memory Usage Demonstration ===\n")
    
    middleware = ConnectionLimiterMiddleware()
    
    print("Memory efficiency features:")
    print("1. Automatic cleanup of empty deques")
    print("2. No recreation of entire data structures")
    print("3. Efficient memory usage patterns")
    print()
    
    ip = "192.168.1.200"
    current_time = time.time()
    old_time = current_time - 120
    
    middleware.connection_attempts[ip] = deque([old_time, old_time, old_time])
    print(f"Before cleanup: {ip} has {len(middleware.connection_attempts[ip])} timestamps")
    
    cutoff_time = current_time - 60
    middleware._cleanup_expired_timestamps(ip, cutoff_time)
    
    exists = ip in middleware.connection_attempts
    print(f"After cleanup: {ip} exists in connection_attempts: {exists}")
    print("✓ Empty deques automatically removed to save memory")


async def integration_demo():
    """Demonstrate optimized middleware in action"""
    
    print("\n=== Integration Demo ===\n")
    
    middleware = ConnectionLimiterMiddleware(
        max_connections_per_ip=50,
        max_connections_per_minute=200,
        sliding_window_size=60,
        cleanup_interval=30
    )
    
    print("Simulating realistic traffic patterns...")
    
    ips = [f"203.0.113.{i}" for i in range(1, 21)]
    
    start_time = time.perf_counter()
    connection_count = 0
    
    for second in range(5):
        for ip in ips:
            attempts = min(15, 5 + (second * 2))
            for _ in range(attempts):
                middleware._record_connection_attempt(ip)
                connection_count += 1
        
        await asyncio.sleep(0.1)
    
    total_time = time.perf_counter() - start_time
    
    print(f"Total connections processed: {connection_count:,}")
    print(f"Processing time: {total_time:.3f}s")
    print(f"Processing rate: {connection_count / total_time:.0f} connections/second")
    
    stats = middleware.get_global_stats()
    print(f"Active connection tracking: {stats['tracked_ips']} IPs")
    print(f"Recent attempts: {stats['total_recent_attempts']}")
    
    print("\n✓ High-performance connection limiting achieved!")


if __name__ == "__main__":
    demonstrate_optimization_benefits()
    performance_comparison()
    demonstrate_memory_usage()
    
    asyncio.run(integration_demo())
    
    print("\n" + "="*60)
    print("Optimization Complete!")
    print("="*60)
    print("The ConnectionLimiterMiddleware now uses collections.deque")
    print("for O(1) cleanup operations and improved performance.")
    print("All rate limiting functionality is preserved while")
    print("significantly improving efficiency under high load.") 


import asyncio
import logging
from typing import Dict, Any, Optional, Deque, Union
import json
from datetime import datetime
import socket
import ssl
import time
from collections import deque
from .types import BaseMessage
from .exceptions import ConnectionError, MessageSizeError, ErrorCategory, ErrorContext, ErrorCategorizer

logger = logging.getLogger(__name__)

DEFAULT_OPERATION_TIMEOUT = 30.0
DEFAULT_MAX_MESSAGE_SIZE = 1024 * 1024

class ErrorMetrics:
    def __init__(self):
        self.connection_errors = 0
        self.timeout_errors = 0
        self.json_errors = 0
        self.size_errors = 0
        self.network_errors = 0
        self.ssl_errors = 0
        self.unexpected_errors = 0
        
        self.fatal_errors = 0
        self.recoverable_errors = 0
        self.client_errors = 0
        self.server_errors = 0
    
    def increment(self, error_type: str):
        if hasattr(self, f"{error_type}_errors"):
            setattr(self, f"{error_type}_errors", getattr(self, f"{error_type}_errors") + 1)
    
    def increment_category(self, category: ErrorCategory):
        category_map = {
            ErrorCategory.FATAL: 'fatal',
            ErrorCategory.RECOVERABLE: 'recoverable', 
            ErrorCategory.CLIENT_ERROR: 'client',
            ErrorCategory.SERVER_ERROR: 'server'
        }
        
        category_name = category_map.get(category, 'server')
        self.increment(category_name)


error_metrics = ErrorMetrics()


class BackpressureMetrics:
    def __init__(self):
        self.messages_queued = 0
        self.messages_sent = 0
        self.messages_dropped = 0
        self.queue_overflow_events = 0
        self.slow_client_disconnections = 0
        self.avg_queue_size = 0.0
        self.max_queue_size_reached = 0
        self.total_send_time_ms = 0.0
        self.send_timeouts = 0
        
        self.connection_metrics: Dict[str, Dict[str, Any]] = {}
        self._start_time = time.time()
    
    def record_message_queued(self, connection_id: str, queue_size: int):
        self.messages_queued += 1
        self._update_connection_metric(connection_id, 'messages_queued', 1)
        self._update_connection_metric(connection_id, 'current_queue_size', queue_size)
        
        if queue_size > self.max_queue_size_reached:
            self.max_queue_size_reached = queue_size
    
    def record_message_sent(self, connection_id: str, send_time_ms: float, queue_size: int):
        self.messages_sent += 1
        self.total_send_time_ms += send_time_ms
        self._update_connection_metric(connection_id, 'messages_sent', 1)
        self._update_connection_metric(connection_id, 'total_send_time_ms', send_time_ms)
        self._update_connection_metric(connection_id, 'current_queue_size', queue_size)
    
    def record_message_dropped(self, connection_id: str, reason: str):
        self.messages_dropped += 1
        self._update_connection_metric(connection_id, 'messages_dropped', 1)
        self._update_connection_metric(connection_id, 'last_drop_reason', reason)
    
    def record_queue_overflow(self, connection_id: str):
        self.queue_overflow_events += 1
        self._update_connection_metric(connection_id, 'queue_overflows', 1)
    
    def record_slow_client_disconnect(self, connection_id: str):
        self.slow_client_disconnections += 1
        self._update_connection_metric(connection_id, 'disconnected_for_slowness', True)
    
    def record_send_timeout(self, connection_id: str):
        self.send_timeouts += 1
        self._update_connection_metric(connection_id, 'send_timeouts', 1)
    
    def _update_connection_metric(self, connection_id: str, metric: str, value: Union[int, float, str, bool]):
        if connection_id not in self.connection_metrics:
            self.connection_metrics[connection_id] = {
                'created_at': time.time(),
                'messages_queued': 0,
                'messages_sent': 0,
                'messages_dropped': 0,
                'queue_overflows': 0,
                'send_timeouts': 0,
                'total_send_time_ms': 0.0,
                'current_queue_size': 0,
                'disconnected_for_slowness': False,
                'last_drop_reason': None
            }
        
        if metric in ['messages_queued', 'messages_sent', 'messages_dropped', 'queue_overflows', 'send_timeouts']:
            self.connection_metrics[connection_id][metric] += value
        elif metric == 'total_send_time_ms':
            self.connection_metrics[connection_id][metric] += value
        else:
            self.connection_metrics[connection_id][metric] = value
    
    def get_connection_stats(self, connection_id: str) -> Dict[str, Any]:
        return self.connection_metrics.get(connection_id, {})
    
    def get_global_stats(self) -> Dict[str, Any]:
        uptime_seconds = time.time() - self._start_time
        avg_send_time = self.total_send_time_ms / max(1, self.messages_sent)
        
        return {
            'uptime_seconds': uptime_seconds,
            'messages_queued': self.messages_queued,
            'messages_sent': self.messages_sent,
            'messages_dropped': self.messages_dropped,
            'queue_overflow_events': self.queue_overflow_events,
            'slow_client_disconnections': self.slow_client_disconnections,
            'max_queue_size_reached': self.max_queue_size_reached,
            'average_send_time_ms': avg_send_time,
            'send_timeouts': self.send_timeouts,
            'active_connections': len(self.connection_metrics),
            'throughput_messages_per_second': self.messages_sent / max(1, uptime_seconds)
        }
    
    def cleanup_connection(self, connection_id: str):
        self.connection_metrics.pop(connection_id, None)


class SendQueueItem:
    def __init__(self, data: str, created_at: float, item_type: str = "data"):
        self.data = data
        self.created_at = created_at
        self.item_type = item_type


class BackpressureManager:
    def __init__(self, connection_id: str, max_queue_size: int = 100, 
                 overflow_strategy: str = "drop_oldest",
                 slow_client_threshold: int = 80,
                 slow_client_timeout: float = 60.0,
                 max_response_time_ms: int = 5000,
                 enable_metrics: bool = True):
        self.connection_id = connection_id
        self.max_queue_size = max_queue_size
        self.overflow_strategy = overflow_strategy
        self.slow_client_threshold = slow_client_threshold
        self.slow_client_timeout = slow_client_timeout
        self.max_response_time_ms = max_response_time_ms
        self.enable_metrics = enable_metrics
        
        self.send_queue: Deque[SendQueueItem] = deque()
        self.is_processing = False
        self.last_successful_send = time.time()
        self.slow_client_detected_at: Optional[float] = None
        
        self._queue_lock = asyncio.Lock()
        self._metrics = BackpressureMetrics() if enable_metrics else None
    
    @property
    def queue_size(self) -> int:
        return len(self.send_queue)
    
    @property
    def queue_utilization_percent(self) -> float:
        return (self.queue_size / max(1, self.max_queue_size)) * 100
    
    @property
    def is_queue_full(self) -> bool:
        return self.queue_size >= self.max_queue_size
    
    @property
    def is_slow_client(self) -> bool:
        threshold_size = int(self.max_queue_size * (self.slow_client_threshold / 100))
        return self.queue_size >= threshold_size
    
    @property
    def should_disconnect_slow_client(self) -> bool:
        if not self.slow_client_detected_at:
            return False
        
        elapsed = time.time() - self.slow_client_detected_at
        return elapsed >= self.slow_client_timeout
    
    async def enqueue_message(self, data: str, item_type: str = "data") -> bool:
        async with self._queue_lock:
            current_time = time.time()
            was_slow_before = self.slow_client_detected_at is not None
            
            if self.is_queue_full:
                if self._metrics:
                    self._metrics.record_queue_overflow(self.connection_id)
                
                if self.overflow_strategy == "drop_oldest":
                    if self.send_queue:
                        dropped_item = self.send_queue.popleft()
                        if self._metrics:
                            self._metrics.record_message_dropped(self.connection_id, "queue_overflow_drop_oldest")
                        logger.debug(f"Dropped oldest message from queue for connection {self.connection_id}")
                elif self.overflow_strategy == "drop_newest":
                    if self._metrics:
                        self._metrics.record_message_dropped(self.connection_id, "queue_overflow_drop_newest")
                    logger.debug(f"Dropped newest message from queue for connection {self.connection_id}")
                    return False
                elif self.overflow_strategy == "reject":
                    if self._metrics:
                        self._metrics.record_message_dropped(self.connection_id, "queue_overflow_reject")
                    logger.warning(f"Message rejected due to full queue for connection {self.connection_id}")
                    return False
            
            queue_item = SendQueueItem(data, current_time, item_type)
            self.send_queue.append(queue_item)
            
            if self.is_slow_client and not was_slow_before:
                self.slow_client_detected_at = current_time
                logger.warning(f"Slow client detected for connection {self.connection_id}, queue size: {self.queue_size}")
            elif not self.is_slow_client and was_slow_before:
                self.slow_client_detected_at = None
            
            if self._metrics:
                self._metrics.record_message_queued(self.connection_id, self.queue_size)
            
            return True
    
    async def dequeue_message(self) -> Optional[SendQueueItem]:
        async with self._queue_lock:
            if self.send_queue:
                return self.send_queue.popleft()
            return None
    
    def record_successful_send(self, send_time_ms: float):
        self.last_successful_send = time.time()
        if self._metrics:
            self._metrics.record_message_sent(self.connection_id, send_time_ms, self.queue_size)
        
        if not self.is_slow_client:
            self.slow_client_detected_at = None
    
    def record_send_timeout(self):
        if self._metrics:
            self._metrics.record_send_timeout(self.connection_id)
    
    def get_health_stats(self) -> Dict[str, Any]:
        current_time = time.time()
        time_since_last_send = current_time - self.last_successful_send
        
        stats = {
            'connection_id': self.connection_id,
            'queue_size': self.queue_size,
            'queue_utilization_percent': self.queue_utilization_percent,
            'is_slow_client': self.is_slow_client,
            'time_since_last_send_seconds': time_since_last_send,
            'slow_client_detected_at': self.slow_client_detected_at,
            'should_disconnect': self.should_disconnect_slow_client,
            'is_processing': self.is_processing
        }
        
        if self._metrics:
            stats.update(self._metrics.get_connection_stats(self.connection_id))
        
        return stats
    
    def cleanup(self):
        if self._metrics:
            self._metrics.cleanup_connection(self.connection_id)


backpressure_metrics = BackpressureMetrics()


class WebSocket:
    def __init__(self, websocket, operation_timeout: float = DEFAULT_OPERATION_TIMEOUT, 
                 max_message_size: int = DEFAULT_MAX_MESSAGE_SIZE,
                 backpressure_settings: Optional[Dict[str, Any]] = None):
        self._websocket = websocket
        self.context: Dict[str, Any] = {}
        
        self._is_closed_event = asyncio.Event()
        self._is_closed_event.clear()
        
        self._send_lock = asyncio.Lock()
        self._receive_lock = asyncio.Lock()
        
        self._close_lock = asyncio.Lock()
        
        self._operation_timeout = operation_timeout
        self._max_message_size = max_message_size
        self._error_count = 0
        
        if max_message_size <= 0:
            raise ValueError("max_message_size must be positive")
        
        self._connection_id = f"ws_{id(self)}_{time.time()}"
        self._backpressure_enabled = False
        self._backpressure_manager: Optional[BackpressureManager] = None
        self._send_task: Optional[asyncio.Task] = None
        
        if backpressure_settings:
            self._init_backpressure(backpressure_settings)
    
    def _init_backpressure(self, settings: Dict[str, Any]):
        self._backpressure_enabled = settings.get('enabled', True)
        
        if self._backpressure_enabled:
            self._backpressure_manager = BackpressureManager(
                connection_id=self._connection_id,
                max_queue_size=settings.get('send_queue_max_size', 100),
                overflow_strategy=settings.get('send_queue_overflow_strategy', 'drop_oldest'),
                slow_client_threshold=settings.get('slow_client_threshold', 80),
                slow_client_timeout=settings.get('slow_client_timeout', 60.0),
                max_response_time_ms=settings.get('max_response_time_ms', 5000),
                enable_metrics=settings.get('enable_send_metrics', True)
            )
            
            self._send_task = None
            self._start_background_worker()
    
    def _start_background_worker(self):
        try:
            if self._send_task is None or self._send_task.done():
                self._send_task = asyncio.create_task(self._background_send_worker())
        except RuntimeError:
            pass
    
    def _ensure_background_worker(self):
        if self._backpressure_enabled and self._backpressure_manager:
            if self._send_task is None or self._send_task.done():
                try:
                    self._send_task = asyncio.create_task(self._background_send_worker())
                except RuntimeError:
                    logger.warning(f"Cannot start background worker for {self._connection_id}: no event loop")
    
    async def _background_send_worker(self):
        if not self._backpressure_manager:
            return
        
        logger.debug(f"Started background send worker for connection {self._connection_id}")
        
        try:
            while not self._is_closed and self._backpressure_enabled:
                try:
                    if self._backpressure_manager.should_disconnect_slow_client:
                        logger.warning(f"Disconnecting slow client {self._connection_id}")
                        backpressure_metrics.record_slow_client_disconnect(self._connection_id)
                        await self.close(code=1008, reason="Slow client detected")
                        break
                    
                    queue_item = await self._backpressure_manager.dequeue_message()
                    if queue_item is None:
                        await asyncio.sleep(0.01)
                        continue
                    
                    start_time = time.time()
                    self._backpressure_manager.is_processing = True
                    
                    try:
                        await asyncio.wait_for(
                            self._websocket.send(queue_item.data),
                            timeout=self._operation_timeout
                        )
                        
                        send_time_ms = (time.time() - start_time) * 1000
                        self._backpressure_manager.record_successful_send(send_time_ms)
                        
                        if send_time_ms > self._backpressure_manager.max_response_time_ms:
                            logger.warning(f"Slow send detected for connection {self._connection_id}: {send_time_ms:.1f}ms")
                        
                    except asyncio.TimeoutError:
                        self._backpressure_manager.record_send_timeout()
                        logger.warning(f"Send timeout for connection {self._connection_id}")
                    
                    except Exception as e:
                        logger.error(f"Send error in background worker for {self._connection_id}: {e}")
                        if "closed" in str(e).lower() or "broken" in str(e).lower():
                            break
                    
                    finally:
                        self._backpressure_manager.is_processing = False
                
                except asyncio.CancelledError:
                    logger.debug(f"Background send worker cancelled for {self._connection_id}")
                    break
                except Exception as e:
                    logger.error(f"Unexpected error in background send worker for {self._connection_id}: {e}")
                    await asyncio.sleep(0.1)
        
        finally:
            logger.debug(f"Background send worker finished for connection {self._connection_id}")
            if self._backpressure_manager:
                self._backpressure_manager.is_processing = False

    @property
    def _is_closed(self) -> bool:
        return self._is_closed_event.is_set()
    
    def _mark_as_closed(self):
        self._is_closed_event.set()
        
        if self._send_task and not self._send_task.done():
            self._send_task.cancel()
        
        if self._backpressure_manager:
            self._backpressure_manager.cleanup()
    
    def _reset_connection_state_for_testing(self):
        self._is_closed_event.clear()
        self._error_count = 0
    
    def _create_error_context(self, operation: str, additional_context: Dict[str, Any] = None) -> ErrorContext:
        context_data = {
            'remote_address': self.remote_address,
            'is_closed': self._is_closed,
            'error_count': self._error_count,
            'operation_timeout': self._operation_timeout,
        }
        
        if additional_context:
            context_data.update(additional_context)
        
        return ErrorContext(
            operation=operation,
            component='websocket',
            additional_context=context_data
        )
    
    def _log_error_with_context(self, error: Exception, context: ErrorContext):
        category = ErrorCategorizer.categorize_exception(error)
        log_level = ErrorCategorizer.get_log_level(error)
        
        log_method = getattr(logger, log_level, logger.error)
        
        message = f"WebSocket {context.operation} error: {error}"
        
        extra_data = {
            'error_category': category.value,
            'error_type': type(error).__name__,
            'error_id': context.error_id,
            'context': context.to_dict()
        }
        
        log_method(message, extra=extra_data)
        
        error_metrics.increment_category(category)
    
    def _handle_critical_error(self, error: Exception, operation: str):
        self._error_count += 1
        self._mark_as_closed()
        
        context = self._create_error_context(operation, {'critical': True})
        self._log_error_with_context(error, context)
    
    async def send_json(self, data: dict) -> None:
        if self._is_closed:
            raise ConnectionError("WebSocket connection is closed")
        
        try:
            def json_serializer(obj):
                if isinstance(obj, datetime):
                    return obj.isoformat()
                raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")
            
            json_data = json.dumps(data, default=json_serializer)
            
            if self._backpressure_enabled and self._backpressure_manager:
                self._ensure_background_worker()
                success = await self._backpressure_manager.enqueue_message(json_data, "json")
                if not success:
                    raise ConnectionError("Failed to queue message: queue overflow")
                return
            
            await self._direct_send(json_data, 'send_json')
            
        except (TypeError, ValueError) as e:
            error_metrics.increment('json')
            context = self._create_error_context('send_json', {'json_serialization': True})
            self._log_error_with_context(e, context)
            raise ConnectionError(f"JSON serialization failed: {str(e)}")
        
        except ConnectionError:
            raise
        
        except Exception as e:
            error_metrics.increment('unexpected')
            self._handle_critical_error(e, 'send_json')
            raise ConnectionError(f"Unexpected error during send: {str(e)}")
    
    async def _direct_send(self, data: str, operation: str) -> None:
        async with self._send_lock:
            if self._is_closed:
                raise ConnectionError("WebSocket connection is closed")
            
            try:
                await self._websocket.send(data)
                self._error_count = 0
                
            except asyncio.TimeoutError as e:
                error_metrics.increment('timeout')
                self._handle_critical_error(e, operation)
                raise ConnectionError(f"Send operation timed out after {self._operation_timeout} seconds")
                
            except asyncio.CancelledError:
                self._mark_as_closed()
                raise
            
            except (socket.error, OSError) as e:
                error_metrics.increment('network')
                self._handle_critical_error(e, operation)
                raise ConnectionError(f"Network error during send: {str(e)}")
            
            except ssl.SSLError as e:
                error_metrics.increment('ssl')
                self._handle_critical_error(e, operation)
                raise ConnectionError(f"SSL error during send: {str(e)}")
            
            except (AttributeError, RuntimeError) as e:
                error_metrics.increment('connection')
                self._handle_critical_error(e, operation)
                raise ConnectionError(f"WebSocket protocol error: {str(e)}")
            
            except Exception as e:
                error_metrics.increment('unexpected')
                self._handle_critical_error(e, operation)
                raise ConnectionError(f"Unexpected error during send: {str(e)}")

    async def send_message(self, message: BaseMessage) -> None:
        await self.send_json(message.dict())
    
    async def receive_json(self) -> dict:
        if self._is_closed:
            raise ConnectionError("WebSocket connection is closed")
        
        async with self._receive_lock:
            if self._is_closed:
                raise ConnectionError("WebSocket connection is closed")
            
            try:
                raw_data = await self._websocket.recv()
                
                if raw_data is None:
                    error_metrics.increment('connection')
                    self._handle_critical_error(Exception("Received None data"), 'receive_json')
                    raise ConnectionError("WebSocket protocol error: received None data")
                
                message_size = len(raw_data)
                if message_size > self._max_message_size:
                    error_metrics.increment('size')
                    context = self._create_error_context('receive_json', {'message_size': message_size})
                    oversized_error = MessageSizeError(f"Message size {message_size} exceeds limit {self._max_message_size}")
                    self._log_error_with_context(oversized_error, context)
                    logger.warning(f"Oversized JSON message blocked: {message_size} bytes (limit: {self._max_message_size})")
                    raise oversized_error
                
                try:
                    result = json.loads(raw_data)
                    self._error_count = 0
                    return result
                except json.JSONDecodeError as e:
                    error_metrics.increment('json')
                    context = self._create_error_context('receive_json', {
                        'message_size': message_size,
                        'raw_data_preview': raw_data[:100] if len(raw_data) > 100 else raw_data
                    })
                    self._log_error_with_context(e, context)
                    raise ConnectionError(f"Invalid JSON received: {str(e)}")
                
            except asyncio.TimeoutError as e:
                error_metrics.increment('timeout')
                self._handle_critical_error(e, 'receive_json')
                raise ConnectionError(f"Receive operation timed out after {self._operation_timeout} seconds")
                
            except asyncio.CancelledError:
                self._mark_as_closed()
                raise
            
            except MessageSizeError:
                raise
            
            except (socket.error, OSError) as e:
                error_metrics.increment('network')
                self._handle_critical_error(e, 'receive_json')
                raise ConnectionError(f"Network error during receive: {str(e)}")
            
            except ssl.SSLError as e:
                error_metrics.increment('ssl')
                self._handle_critical_error(e, 'receive_json')
                raise ConnectionError(f"SSL error during receive: {str(e)}")
            
            except (AttributeError, RuntimeError) as e:
                error_metrics.increment('connection')
                self._handle_critical_error(e, 'receive_json')
                raise ConnectionError(f"WebSocket protocol error: {str(e)}")
            
            except Exception as e:
                error_metrics.increment('unexpected')
                self._handle_critical_error(e, 'receive_json')
                raise ConnectionError(f"Unexpected error during receive: {str(e)}")
    
    async def recv(self) -> str:
        if self._is_closed:
            raise ConnectionError("WebSocket connection is closed")
        
        async with self._receive_lock:
            if self._is_closed:
                raise ConnectionError("WebSocket connection is closed")
            
            try:
                raw_data = await self._websocket.recv()
                
                if raw_data is None:
                    error_metrics.increment('connection')
                    self._handle_critical_error(Exception("Received None data"), 'recv')
                    raise ConnectionError("WebSocket protocol error: received None data")
                
                message_size = len(raw_data)
                if message_size > self._max_message_size:
                    error_metrics.increment('size')
                    context = self._create_error_context('recv', {'message_size': message_size})
                    oversized_error = MessageSizeError(f"Message size {message_size} exceeds limit {self._max_message_size}")
                    self._log_error_with_context(oversized_error, context)
                    logger.warning(f"Oversized message blocked: {message_size} bytes (limit: {self._max_message_size})")
                    raise oversized_error
                
                self._error_count = 0
                return raw_data
                
            except asyncio.TimeoutError as e:
                error_metrics.increment('timeout')
                self._handle_critical_error(e, 'recv')
                raise ConnectionError(f"Receive operation timed out after {self._operation_timeout} seconds")
                
            except asyncio.CancelledError:
                self._mark_as_closed()
                raise
            
            except MessageSizeError:
                raise
            
            except ssl.SSLError as e:
                error_metrics.increment('ssl')
                self._handle_critical_error(e, 'recv')
                raise ConnectionError(f"SSL error during receive: {str(e)}")
            
            except (socket.error, OSError) as e:
                error_metrics.increment('network')
                self._handle_critical_error(e, 'recv')
                raise ConnectionError(f"Network error during receive: {str(e)}")
            
            except (AttributeError, RuntimeError) as e:
                error_metrics.increment('connection')
                self._handle_critical_error(e, 'recv')
                raise ConnectionError(f"WebSocket protocol error: {str(e)}")
            
            except Exception as e:
                error_metrics.increment('unexpected')
                self._handle_critical_error(e, 'recv')
                raise ConnectionError(f"Unexpected error during receive: {str(e)}")
    
    async def send(self, data: str) -> None:
        if self._is_closed:
            raise ConnectionError("WebSocket connection is closed")
        
        if self._backpressure_enabled and self._backpressure_manager:
            self._ensure_background_worker()
            success = await self._backpressure_manager.enqueue_message(data, "data")
            if not success:
                raise ConnectionError("Failed to queue message: queue overflow")
            return
        
        await self._direct_send(data, 'send')
    
    async def close(self, code: int = 1000, reason: str = "") -> None:
        async with self._close_lock:
            if self._is_closed:
                logger.debug("WebSocket connection already closed, ignoring close() call")
                return
            
            self._mark_as_closed()
            
            try:
                await asyncio.wait_for(
                    self._websocket.close(code=code, reason=reason),
                    timeout=self._operation_timeout
                )
                logger.debug(f"WebSocket connection closed gracefully with code {code}")
                
            except asyncio.TimeoutError as e:
                error_metrics.increment('timeout')
                context = self._create_error_context('close', {'timeout': True, 'code': code})
                self._log_error_with_context(e, context)
                logger.warning(f"WebSocket close operation timed out after {self._operation_timeout} seconds")
            
            except asyncio.CancelledError:
                logger.debug("WebSocket close operation was cancelled")
                raise
            
            except (socket.error, OSError) as e:
                context = self._create_error_context('close', {'network_error': True, 'code': code})
                self._log_error_with_context(e, context)
                logger.debug(f"Network error during WebSocket close: {str(e)}")
            
            except ssl.SSLError as e:
                context = self._create_error_context('close', {'ssl_error': True, 'code': code})
                self._log_error_with_context(e, context)
                logger.debug(f"SSL error during WebSocket close: {str(e)}")
            
            except (AttributeError, RuntimeError) as e:
                context = self._create_error_context('close', {'protocol_error': True, 'code': code})
                self._log_error_with_context(e, context)
                logger.debug(f"Protocol error during WebSocket close: {str(e)}")
            
            except Exception as e:
                context = self._create_error_context('close', {'unexpected': True, 'code': code})
                self._log_error_with_context(e, context)
                logger.warning(f"Unexpected error during WebSocket close: {type(e).__name__}: {e}")
    
    @property
    def closed(self) -> bool:
        return self._is_closed
    
    @property  
    def is_closed(self) -> bool:
        return self._is_closed
    
    @property
    def remote_address(self) -> tuple:
        try:
            return getattr(self._websocket, 'remote_address', ('unknown', 0))
        except (AttributeError, OSError, RuntimeError) as e:
            logger.debug(f"Could not get remote address: {type(e).__name__}: {e}")
            return ('unknown', 0)
        except Exception as e:
            logger.warning(f"Unexpected error getting remote address: {type(e).__name__}: {e}")
            return ('unknown', 0)
    
    def set_operation_timeout(self, timeout: float) -> None:
        if timeout <= 0:
            raise ValueError("Timeout must be positive")
        self._operation_timeout = timeout
    
    def enable_backpressure(self, settings: Dict[str, Any]) -> None:
        if not self._backpressure_enabled:
            self._init_backpressure(settings)
    
    def disable_backpressure(self) -> None:
        self._backpressure_enabled = False
        
        if self._send_task and not self._send_task.done():
            self._send_task.cancel()
        
        if self._backpressure_manager:
            self._backpressure_manager.cleanup()
            self._backpressure_manager = None
    
    def get_backpressure_stats(self) -> Dict[str, Any]:
        if not self._backpressure_manager:
            return {
                'backpressure_enabled': False,
                'connection_id': self._connection_id
            }
        
        stats = self._backpressure_manager.get_health_stats()
        stats['backpressure_enabled'] = self._backpressure_enabled
        return stats
    
    @property
    def connection_id(self) -> str:
        return self._connection_id
    
    @property
    def backpressure_enabled(self) -> bool:
        return self._backpressure_enabled
    
    @property
    def send_queue_size(self) -> int:
        if self._backpressure_manager:
            return self._backpressure_manager.queue_size
        return 0
    
    @property
    def send_queue_utilization_percent(self) -> float:
        if self._backpressure_manager:
            return self._backpressure_manager.queue_utilization_percent
        return 0.0
        
    @property
    def error_metrics(self) -> ErrorMetrics:
        return error_metrics
    
    @property
    def backpressure_metrics(self) -> BackpressureMetrics:
        return backpressure_metrics 


"""
Event dispatcher implementation
"""

import asyncio
import logging
import threading
import time
from .router import Router
from .websocket import WebSocket  
from .types import BaseMessage, ChatMessage, JoinRoomMessage, GameActionMessage, EventType
from .exceptions import (
    MessageValidationError, MiddlewareError, ConnectionError, AiowsException,
    ErrorCategory, ErrorContext, ErrorCategorizer
)
from .middleware.base import BaseMiddleware
from typing import List, Optional, Dict, Any


class DispatcherErrorMetrics:
    """Error metrics collector for dispatcher operations"""
    def __init__(self):
        self.middleware_errors = 0
        self.handler_errors = 0
        self.parsing_errors = 0
        self.critical_errors = 0
        self.timeout_errors = 0
        self.validation_errors = 0
        self.connection_errors = 0
        self.unexpected_errors = 0
        self.last_error_time = 0
        
        self.fatal_errors = 0
        self.recoverable_errors = 0
        self.client_errors = 0
        self.server_errors = 0
    
    def increment(self, error_type: str):
        if hasattr(self, f"{error_type}_errors"):
            setattr(self, f"{error_type}_errors", getattr(self, f"{error_type}_errors") + 1)
        self.last_error_time = time.time()
    
    def increment_category(self, category: ErrorCategory):
        category_map = {
            ErrorCategory.FATAL: 'fatal',
            ErrorCategory.RECOVERABLE: 'recoverable', 
            ErrorCategory.CLIENT_ERROR: 'client',
            ErrorCategory.SERVER_ERROR: 'server'
        }
        
        category_name = category_map.get(category, 'server')
        self.increment(category_name)


dispatcher_error_metrics = DispatcherErrorMetrics()


class MessageDispatcher:
    """Dispatcher for handling WebSocket events and messages"""
    
    def __init__(self, router: Router):
        self.router = router
        self._middleware: List[BaseMiddleware] = []
        self._middleware_lock = threading.Lock()
        self.logger = logging.getLogger("aiows.dispatcher")
        self._consecutive_errors = 0
    
    def add_middleware(self, middleware: BaseMiddleware) -> None:
        with self._middleware_lock:
            self._middleware.append(middleware)
    
    def remove_middleware(self, middleware: BaseMiddleware) -> bool:
        with self._middleware_lock:
            try:
                self._middleware.remove(middleware)
                return True
            except ValueError:
                return False
    
    def _create_error_context(self, operation: str, websocket: Optional[WebSocket] = None, 
                             additional_context: Optional[Dict[str, Any]] = None) -> ErrorContext:
        context_data = {
            'consecutive_errors': self._consecutive_errors,
            'timestamp': time.time(),
        }
        
        if websocket:
            context_data.update({
                'remote_address': websocket.remote_address,
                'websocket_closed': websocket.closed,
            })
        
        if additional_context:
            context_data.update(additional_context)
        
        return ErrorContext(
            operation=operation,
            component='dispatcher',
            additional_context=context_data
        )
    
    def _log_error_with_context(self, error: Exception, context: ErrorContext):
        category = ErrorCategorizer.categorize_exception(error)
        log_level = ErrorCategorizer.get_log_level(error)
        
        log_method = getattr(self.logger, log_level, self.logger.error)
        
        message = f"Dispatcher error in {context.operation}: {error}"
        
        extra_data = {
            'error_category': category.value,
            'error_type': type(error).__name__,
            'error_id': context.error_id,
            'context': context.to_dict()
        }
        
        log_method(message, extra=extra_data)
        
        dispatcher_error_metrics.increment_category(category)
    
    def _parse_message_safely(self, message_data: dict) -> BaseMessage:
        context = self._create_error_context('message_parsing', 
                                           additional_context={'message_type': message_data.get('type', 'unknown')})
        
        try:
            message_type = message_data.get('type')
            
            if message_type == 'chat':
                return ChatMessage(**message_data)
            elif message_type == 'join_room':
                return JoinRoomMessage(**message_data)
            elif message_type == 'game_action':
                return GameActionMessage(**message_data)
            else:
                return BaseMessage(**message_data)
                
        except (TypeError, ValueError, KeyError) as e:
            dispatcher_error_metrics.increment('parsing')
            validation_error = MessageValidationError(f"Failed to parse message: {str(e)}")
            self._log_error_with_context(validation_error, context)
            raise validation_error
        
        except RecursionError as e:
            dispatcher_error_metrics.increment('parsing')
            context.additional_context['recursion_error'] = True
            validation_error = MessageValidationError(f"Message structure too complex: {str(e)}")
            self._log_error_with_context(validation_error, context)
            raise validation_error
        
        except MemoryError as e:
            dispatcher_error_metrics.increment('critical')
            self._log_error_with_context(e, context)
            raise
        
        except Exception as e:
            dispatcher_error_metrics.increment('unexpected')
            context.additional_context['unexpected'] = True
            self._log_error_with_context(e, context)
            self.logger.critical(f"Unexpected error during message parsing: {type(e).__name__}: {e}")
            raise MessageValidationError(f"Unexpected parsing error: {str(e)}")
    
    async def _handle_middleware_exception(
        self, 
        exception: Exception, 
        middleware: BaseMiddleware, 
        event_type: str, 
        websocket: Optional[WebSocket] = None
    ) -> bool:
        middleware_name = middleware.__class__.__name__
        
        context = self._create_error_context(
            f"middleware_{event_type}",
            websocket,
            {'middleware_name': middleware_name}
        )
        
        self._consecutive_errors += 1
        
        category = ErrorCategorizer.categorize_exception(exception)
        should_stop = ErrorCategorizer.should_stop_middleware_chain(exception, middleware_name)
        
        self._log_error_with_context(exception, context)
        
        if category == ErrorCategory.FATAL:
            if isinstance(exception, (MemoryError, OSError)):
                dispatcher_error_metrics.increment('critical')
                self.logger.critical(f"System error in {middleware_name} during {event_type}: {str(exception)}")
            return False
        
        elif category == ErrorCategory.CLIENT_ERROR:
            return False
            
        elif category == ErrorCategory.RECOVERABLE:
            return False
            
        elif category == ErrorCategory.SERVER_ERROR:
            if isinstance(exception, ConnectionError):
                return False
            elif isinstance(exception, MiddlewareError):
                if "auth" in middleware_name.lower() or "security" in middleware_name.lower():
                    self.logger.warning(f"Critical middleware {middleware_name} failed, interrupting chain")
                    return False
                return True
            else:
                return True
        
        return not should_stop
    
    async def _execute_handler_chain_safely(self, handler_type: str, handler_list: List, *args) -> None:
        websocket = args[0] if args and hasattr(args[0], 'context') else None
        
        for handler in handler_list:
            try:
                await handler(*args)
                self._consecutive_errors = max(0, self._consecutive_errors - 1)
                
            except (asyncio.CancelledError, asyncio.TimeoutError) as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('timeout')
                context = self._create_error_context(f'{handler_type}_handler', websocket)
                self._log_error_with_context(e, context)
                raise
            
            except (MemoryError, OSError) as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('critical')
                context = self._create_error_context(f'{handler_type}_handler', websocket, {'critical': True})
                self._log_error_with_context(e, context)
                raise
            
            except (ConnectionError, MessageValidationError) as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('handler')
                context = self._create_error_context(f'{handler_type}_handler', websocket)
                self._log_error_with_context(e, context)
                
            except Exception as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('unexpected')
                context = self._create_error_context(f'{handler_type}_handler', websocket, {'unexpected': True})
                self._log_error_with_context(e, context)
                self.logger.warning(f"Unexpected error in {handler_type} handler: {type(e).__name__}: {e}")
    
    async def _execute_connect_chain(self, websocket: WebSocket) -> None:
        await self._execute_handler_chain_safely('connect', self.router._connect_handlers, websocket)
    
    async def _execute_disconnect_chain(self, websocket: WebSocket, reason: str) -> None:
        await self._execute_handler_chain_safely('disconnect', self.router._disconnect_handlers, websocket, reason)
    
    async def _execute_message_chain(self, websocket: WebSocket, message_data: dict) -> None:
        message = self._parse_message_safely(message_data)
        message_type = message.type
        
        suitable_handler = None
        
        for handler_info in self.router._message_handlers:
            handler_message_type = handler_info.get('message_type')
            
            if handler_message_type is None or handler_message_type == message_type:
                suitable_handler = handler_info.get('handler')
                break
        
        if suitable_handler:
            try:
                await suitable_handler(websocket, message)
                self._consecutive_errors = max(0, self._consecutive_errors - 1)
                
            except (asyncio.CancelledError, asyncio.TimeoutError) as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('timeout')
                context = self._create_error_context('message_handler', websocket, 
                                                   {'message_type': message_type})
                self._log_error_with_context(e, context)
                raise
            
            except (MemoryError, OSError) as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('critical')
                context = self._create_error_context('message_handler', websocket, 
                                                   {'critical': True, 'message_type': message_type})
                self._log_error_with_context(e, context)
                raise
            
            except (ConnectionError, MessageValidationError) as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('handler')
                context = self._create_error_context('message_handler', websocket, 
                                                   {'message_type': message_type})
                self._log_error_with_context(e, context)
                
            except Exception as e:
                self._consecutive_errors += 1
                dispatcher_error_metrics.increment('unexpected')
                context = self._create_error_context('message_handler', websocket, 
                                                   {'unexpected': True, 'message_type': message_type})
                self._log_error_with_context(e, context)
                self.logger.warning(f"Unexpected error in message handler: {type(e).__name__}: {e}")
        else:
            self.logger.warning(f"No handler found for message type: {message_type}")

    async def _execute_middleware_chain(self, event_type: EventType, middleware_list: List[BaseMiddleware], *args) -> None:
        middleware_index = 0
        chain_interrupted = False
        
        async def next_handler(*handler_args) -> None:
            nonlocal middleware_index, chain_interrupted
            
            effective_args = handler_args if handler_args else args
            
            if middleware_index < len(middleware_list):
                current_middleware = middleware_list[middleware_index]
                middleware_index += 1
                
                try:
                    if event_type == EventType.CONNECT:
                        await current_middleware.on_connect(next_handler, *effective_args)
                    elif event_type == EventType.DISCONNECT:
                        await current_middleware.on_disconnect(next_handler, *effective_args)
                    elif event_type == EventType.MESSAGE:
                        await current_middleware.on_message(next_handler, *effective_args)
                    else:
                        raise ValueError(f"Unknown event type: {event_type}")
                        
                except Exception as e:
                    websocket = effective_args[0] if effective_args and hasattr(effective_args[0], 'context') else None
                    
                    should_continue = await self._handle_middleware_exception(
                        e, current_middleware, event_type.value, websocket
                    )
                    
                    if not should_continue:
                        chain_interrupted = True
                        if event_type == EventType.CONNECT and websocket and not websocket.closed:
                            try:
                                await websocket.close(code=1011, reason="Server error")
                            except Exception:
                                pass
                        return
                    
                    category = ErrorCategorizer.categorize_exception(e)
                    if category in [ErrorCategory.CLIENT_ERROR, ErrorCategory.RECOVERABLE]:
                        chain_interrupted = True
                        return
                    elif category == ErrorCategory.SERVER_ERROR and isinstance(e, MiddlewareError):
                        middleware_name = current_middleware.__class__.__name__
                        if any(keyword in middleware_name.lower() for keyword in ['auth', 'security']):
                            chain_interrupted = True
                            return
                        await next_handler(*effective_args)
                        return
                    else:
                        chain_interrupted = True
                        return
            else:
                if not chain_interrupted:
                    if event_type == EventType.CONNECT:
                        await self._execute_connect_chain(*effective_args)
                    elif event_type == EventType.DISCONNECT:
                        await self._execute_disconnect_chain(*effective_args)
                    elif event_type == EventType.MESSAGE:
                        await self._execute_message_chain(*effective_args)
        
        await next_handler(*args)

    async def dispatch_connect(self, websocket: WebSocket) -> None:
        with self._middleware_lock:
            middleware_snapshot = tuple(self._middleware)
        
        await self._execute_middleware_chain(EventType.CONNECT, middleware_snapshot, websocket)

    async def dispatch_disconnect(self, websocket: WebSocket, reason: str) -> None:
        with self._middleware_lock:
            middleware_snapshot = tuple(self._middleware)
        
        await self._execute_middleware_chain(EventType.DISCONNECT, middleware_snapshot, websocket, reason)

    async def dispatch_message(self, websocket: WebSocket, message_data: dict) -> None:
        with self._middleware_lock:
            middleware_snapshot = tuple(self._middleware)
        
        await self._execute_middleware_chain(EventType.MESSAGE, middleware_snapshot, websocket, message_data)
    
    @property
    def error_metrics(self) -> DispatcherErrorMetrics:
        return dispatcher_error_metrics



"""
aiows - Modern WebSocket framework inspired by aiogram
"""

from .server import WebSocketServer
from .router import Router
from .websocket import WebSocket
from .dispatcher import MessageDispatcher
from .types import BaseMessage, ChatMessage, JoinRoomMessage, GameActionMessage
from .exceptions import AiowsException, ConnectionError, MessageValidationError
from .middleware import BaseMiddleware, AuthMiddleware, LoggingMiddleware, RateLimitingMiddleware, ConnectionLimiterMiddleware
from .health import (
    HealthStatus, HealthCheck, HealthChecker,
    ConnectionsHealthCheck, MiddlewareHealthCheck, MemoryHealthCheck, SystemHealthCheck,
    setup_health_checks, get_health_checker
)

__all__ = [
    # Core components
    "WebSocketServer",
    "Router", 
    "WebSocket",
    "MessageDispatcher",
    
    # Message types
    "BaseMessage",
    "ChatMessage",
    "JoinRoomMessage", 
    "GameActionMessage",
    
    # Exceptions
    "AiowsException",
    "ConnectionError",
    "MessageValidationError",
    
    # Middleware
    "BaseMiddleware",
    "AuthMiddleware",
    "LoggingMiddleware", 
    "RateLimitingMiddleware",
    "ConnectionLimiterMiddleware",
    
    # Health monitoring
    "HealthStatus",
    "HealthCheck", 
    "HealthChecker",
    "ConnectionsHealthCheck",
    "MiddlewareHealthCheck",
    "MemoryHealthCheck",
    "SystemHealthCheck",
    "setup_health_checks",
    "get_health_checker"
]

__version__ = "0.1.2" 


"""
Example demonstrating backpressure handling in aiows WebSocket framework.

This example shows:
- How to configure backpressure settings
- How to monitor connection health
- How slow clients are automatically handled
- How to access backpressure metrics
"""

import asyncio
import json
import time
from aiows import WebSocketServer, Router
from aiows.settings import create_settings


async def main():
    print("=== aiows Backpressure Handling Example ===\n")
    
    router = Router()
    
    @router.on_connect
    async def handle_connect(websocket):
        print(f"Client connected: {websocket.connection_id}")
        print(f"Backpressure enabled: {websocket.backpressure_enabled}")
        
        if websocket.backpressure_enabled:
            stats = websocket.get_backpressure_stats()
            print(f"Initial queue size: {stats['queue_size']}")
            print(f"Queue utilization: {stats['queue_utilization_percent']:.1f}%")
        
        await websocket.send_json({
            "type": "welcome",
            "message": "Connected to aiows server with backpressure protection!",
            "connection_id": websocket.connection_id
        })
    
    @router.on_message("chat")
    async def handle_chat(websocket, message):
        print(f"Received chat from {websocket.connection_id}: {message.get('text', '')}")
        
        broadcast_msg = {
            "type": "chat_broadcast",
            "sender": websocket.connection_id,
            "text": message.get('text', ''),
            "timestamp": time.time()
        }
        
        for i in range(5):
            await websocket.send_json({
                **broadcast_msg,
                "sequence": i
            })
    
    @router.on_message("simulate_slow_client")
    async def handle_slow_client_simulation(websocket, message):
        print(f"Simulating slow client for {websocket.connection_id}")
        
        for i in range(50):
            await websocket.send_json({
                "type": "bulk_message",
                "sequence": i,
                "data": f"Message {i} of bulk send"
            })
            
            if i % 10 == 0:
                stats = websocket.get_backpressure_stats()
                print(f"Queue size: {stats['queue_size']}, "
                      f"Utilization: {stats['queue_utilization_percent']:.1f}%, "
                      f"Is slow client: {stats['is_slow_client']}")
    
    @router.on_message("get_stats")
    async def handle_get_stats(websocket, message):
        if websocket.backpressure_enabled:
            stats = websocket.get_backpressure_stats()
            await websocket.send_json({
                "type": "stats_response",
                "stats": stats
            })
        else:
            await websocket.send_json({
                "type": "stats_response",
                "error": "Backpressure not enabled"
            })
    
    @router.on_disconnect
    async def handle_disconnect(websocket, reason):
        print(f"Client disconnected: {websocket.connection_id}, reason: {reason}")
        
        if websocket.backpressure_enabled:
            final_stats = websocket.get_backpressure_stats()
            print(f"Final stats - Queue size: {final_stats['queue_size']}, "
                  f"Was slow client: {final_stats['is_slow_client']}")
    
    settings = create_settings("development")
    
    settings.backpressure.enabled = True
    settings.backpressure.send_queue_max_size = 20
    settings.backpressure.send_queue_overflow_strategy = "drop_oldest"
    settings.backpressure.slow_client_threshold = 75
    settings.backpressure.slow_client_timeout = 10.0
    settings.backpressure.max_response_time_ms = 2000
    settings.backpressure.enable_send_metrics = True
    
    print("Backpressure Configuration:")
    print(f"- Queue max size: {settings.backpressure.send_queue_max_size}")
    print(f"- Overflow strategy: {settings.backpressure.send_queue_overflow_strategy}")
    print(f"- Slow client threshold: {settings.backpressure.slow_client_threshold}%")
    print(f"- Slow client timeout: {settings.backpressure.slow_client_timeout}s")
    print(f"- Max response time: {settings.backpressure.max_response_time_ms}ms")
    print()
    
    server = WebSocketServer.from_settings(settings)
    server.include_router(router)
    
    print("Starting server on ws://localhost:8000")
    print("\nTo test backpressure:")
    print("1. Connect with a WebSocket client")
    print("2. Send: {\"type\": \"chat\", \"text\": \"Hello!\"}")
    print("3. Send: {\"type\": \"simulate_slow_client\"}")
    print("4. Send: {\"type\": \"get_stats\"}")
    print("\nPress Ctrl+C to stop")
    print()
    
    async def monitor_server_stats():
        while True:
            await asyncio.sleep(5)
            try:
                backpressure_stats = server.get_backpressure_stats()
                global_stats = backpressure_stats['global_stats']
                
                if global_stats['active_connections'] > 0:
                    print(f"\n=== Server Stats ===")
                    print(f"Active connections: {global_stats['active_connections']}")
                    print(f"Messages queued: {global_stats['messages_queued']}")
                    print(f"Messages sent: {global_stats['messages_sent']}")
                    print(f"Messages dropped: {global_stats['messages_dropped']}")
                    print(f"Slow client disconnections: {global_stats['slow_client_disconnections']}")
                    print(f"Average send time: {global_stats['average_send_time_ms']:.2f}ms")
                    print(f"Throughput: {global_stats['throughput_messages_per_second']:.2f} msg/s")
                    
                    slow_connections = server.get_slow_connections()
                    if slow_connections:
                        print(f"Slow connections: {len(slow_connections)}")
                        for conn in slow_connections:
                            print(f"  - {conn['connection_id']}: queue {conn['queue_size']}/{conn.get('max_queue_size', 'unknown')}")
                    print()
            except Exception as e:
                print(f"Error monitoring stats: {e}")
    
    monitor_task = asyncio.create_task(monitor_server_stats())
    
    try:
        await server.serve("localhost", 8000)
    except KeyboardInterrupt:
        print("\nShutting down server...")
        monitor_task.cancel()
        await server.shutdown()
        
        final_stats = server.get_backpressure_stats()
        print("\n=== Final Statistics ===")
        print(json.dumps(final_stats['global_stats'], indent=2))


if __name__ == "__main__":
    asyncio.run(main()) 


"""
Comprehensive input validation system for aiows
Provides protection against SQL injection, XSS, command injection, path traversal and other attacks
"""

import re
from typing import Any


class SecurityLimits:
    MAX_TEXT_LENGTH = 10000
    MAX_USERNAME_LENGTH = 50
    MAX_ROOM_ID_LENGTH = 100
    MAX_ACTION_LENGTH = 50
    MAX_JSON_DEPTH = 10
    MAX_JSON_SIZE = 1048576  # 1MB
    MAX_ARRAY_LENGTH = 1000
    MAX_OBJECT_KEYS = 100


class SecurityPatterns:
    SQL_INJECTION_PATTERNS = [
        r"(?i)(union\s+select\s+|select\s+.*\s+from\s+|insert\s+into\s+|update\s+.*\s+set\s+|delete\s+from\s+)",
        r"(?i)(drop\s+table\s+|alter\s+table\s+|create\s+table\s+)",
        r"(?i)(exec\s*\(|execute\s*\(|sp_executesql)",
        r"(?i)(\-\-[^\n]*|\/\*.*?\*\/)",
        r"(?i)(\s+(or|and)\s+1\s*=\s*1\s*)",
        r"(?i)(0x[0-9a-f]+|char\s*\(|ascii\s*\()",
        r"[';]\s*(\-\-|drop|delete|insert|update|union|select)",
        r"(?i)(\'\s*(or|and|union|select|drop|delete|insert|update)\s*)",
    ]
    
    XSS_PATTERNS = [
        r"(?i)<script[^>]*>.*?</script>",
        r"(?i)<iframe[^>]*>.*?</iframe>",
        r"(?i)<object[^>]*>.*?</object>",
        r"(?i)<embed[^>]*>.*?</embed>",
        r"(?i)<link[^>]*>",
        r"(?i)<meta[^>]*>",
        r"(?i)javascript:",
        r"(?i)vbscript:",
        r"(?i)data:text/html",
        r"(?i)on(load|click|error|focus|blur|change|submit|keypress|keydown|keyup|mouseover|mouseout)\s*=",
        r"(?i)expression\s*\(",
        r"(?i)<\s*\/?\s*(script|iframe|object|embed|link|meta|style|form|input|button|textarea|select|option)",
    ]
    
    COMMAND_INJECTION_PATTERNS = [
        r"[;&|`$(){}]",
        r"(?i)(bash|sh|cmd|powershell|pwsh)",
        r"(?i)(wget|curl|nc|netcat)",
        r"(?i)(rm\s|del\s|format\s)",
        r"(?i)(cat\s|type\s|more\s|less\s)",
        r"(?i)(echo\s|print\s)",
        r"(?i)(chmod\s|chown\s|sudo\s)",
    ]
    
    PATH_TRAVERSAL_PATTERNS = [
        r"\.\.\/",
        r"\.\.\\",
        r"%2e%2e%2f",
        r"%2e%2e%5c",
        r"\.\.%2f",
        r"\.\.%5c",
        r"%252e%252e%252f",
        r"(?i)(\/etc\/|\/bin\/|\/usr\/|\/var\/|\/tmp\/|\/home\/)",
        r"(?i)(c:\\|d:\\|\\windows\\|\\system32\\)",
        r"\\\\[^\\]+\\[^\\]+",
        r"\.{4,}[\/\\]",
    ]
    
    DANGEROUS_PATTERNS = [
        r"(?i)(eval\s*\(|function\s*\()",
        r"(?i)(import\s+|require\s*\(|include\s*\()",
        r"(?i)(file_get_contents|readfile|fopen|fread)",
        r"(?i)(system\s*\(|shell_exec\s*\(|exec\s*\(|passthru\s*\()",
        r"(?i)(base64_decode|urldecode|rawurldecode)",
    ]


class SecurityValidator:
    @staticmethod
    def check_sql_injection(value: str) -> bool:
        for pattern in SecurityPatterns.SQL_INJECTION_PATTERNS:
            if re.search(pattern, value):
                return True
        return False
    
    @staticmethod
    def check_xss(value: str) -> bool:
        for pattern in SecurityPatterns.XSS_PATTERNS:
            if re.search(pattern, value):
                return True
        return False
    
    @staticmethod
    def check_command_injection(value: str) -> bool:
        for pattern in SecurityPatterns.COMMAND_INJECTION_PATTERNS:
            if re.search(pattern, value):
                return True
        return False
    
    @staticmethod
    def check_path_traversal(value: str) -> bool:
        for pattern in SecurityPatterns.PATH_TRAVERSAL_PATTERNS:
            if re.search(pattern, value):
                return True
        return False
    
    @staticmethod
    def check_dangerous_patterns(value: str) -> bool:
        for pattern in SecurityPatterns.DANGEROUS_PATTERNS:
            if re.search(pattern, value):
                return True
        return False
    
    @staticmethod
    def is_safe_string(value: str) -> bool:
        return not (
            SecurityValidator.check_sql_injection(value) or
            SecurityValidator.check_xss(value) or
            SecurityValidator.check_command_injection(value) or
            SecurityValidator.check_path_traversal(value) or
            SecurityValidator.check_dangerous_patterns(value)
        )


class Sanitizer:
    REMOVE_CHARS = set(['<', '>', '"', "'", '&', '\x00', '\r'])
    
    ESCAPE_MAP = {
        '&': '&amp;',
        '<': '&lt;',
        '>': '&gt;',
        '"': '&quot;',
        "'": '&#x27;',
    }
    
    @staticmethod
    def sanitize_text(value: str) -> str:
        if not isinstance(value, str):
            return str(value)
        
        sanitized = ''.join(char for char in value if char not in Sanitizer.REMOVE_CHARS)
        sanitized = ' '.join(sanitized.split())
        
        return sanitized
    
    @staticmethod
    def sanitize_identifier(value: str) -> str:
        if not isinstance(value, str):
            return str(value)
        
        sanitized = ''.join(char for char in value if char not in Sanitizer.REMOVE_CHARS)
        sanitized = re.sub(r'[^a-zA-Z0-9_.-]', '', sanitized)
        sanitized = re.sub(r'^[.-]+', '', sanitized)
        
        return sanitized
    
    @staticmethod
    def sanitize_action(value: str) -> str:
        if not isinstance(value, str):
            return str(value)
        
        sanitized = re.sub(r'[^a-zA-Z0-9_]', '', value)
        
        return sanitized


class JSONBombProtector:
    @staticmethod
    def check_json_bomb(data: Any, depth: int = 0) -> bool:
        if depth > SecurityLimits.MAX_JSON_DEPTH:
            return True
        
        if isinstance(data, dict):
            if len(data) > SecurityLimits.MAX_OBJECT_KEYS:
                return True
            for value in data.values():
                if JSONBombProtector.check_json_bomb(value, depth + 1):
                    return True
        
        elif isinstance(data, list):
            if len(data) > SecurityLimits.MAX_ARRAY_LENGTH:
                return True
            for item in data:
                if JSONBombProtector.check_json_bomb(item, depth + 1):
                    return True
        
        elif isinstance(data, str):
            if len(data) > SecurityLimits.MAX_TEXT_LENGTH:
                return True
        
        return False
    
    @staticmethod
    def validate_json_size(json_string: str) -> bool:
        return len(json_string.encode('utf-8')) <= SecurityLimits.MAX_JSON_SIZE


class WhitelistValidator:
    USERNAME_CHARS = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-')
    ROOM_ID_CHARS = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-')
    ACTION_CHARS = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_')
    
    ALLOWED_ACTIONS = {
        'move', 'attack', 'defend', 'use_item', 'cast_spell', 
        'jump', 'run', 'walk', 'look', 'take', 'drop',
        'open', 'close', 'push', 'pull', 'activate'
    }
    
    @staticmethod
    def validate_username(value: str) -> bool:
        if not value or len(value) > SecurityLimits.MAX_USERNAME_LENGTH:
            return False
        return all(char in WhitelistValidator.USERNAME_CHARS for char in value)
    
    @staticmethod
    def validate_room_id(value: str) -> bool:
        if not value or len(value) > SecurityLimits.MAX_ROOM_ID_LENGTH:
            return False
        return all(char in WhitelistValidator.ROOM_ID_CHARS for char in value)
    
    @staticmethod
    def validate_action(value: str) -> bool:
        if not value or len(value) > SecurityLimits.MAX_ACTION_LENGTH:
            return False
        return (all(char in WhitelistValidator.ACTION_CHARS for char in value) and
                value.lower() in WhitelistValidator.ALLOWED_ACTIONS)


def validate_safe_text(cls, value: str) -> str:
    if not isinstance(value, str):
        raise ValueError("Value must be a string")
    
    if len(value) > SecurityLimits.MAX_TEXT_LENGTH:
        raise ValueError(f"Text too long: maximum {SecurityLimits.MAX_TEXT_LENGTH} characters")
    
    if not SecurityValidator.is_safe_string(value):
        raise ValueError("Text contains dangerous patterns")
    
    return Sanitizer.sanitize_text(value)


def validate_username(cls, value: str) -> str:
    if not isinstance(value, str):
        raise ValueError("Username must be a string")
    
    sanitized = Sanitizer.sanitize_identifier(value)
    
    if not WhitelistValidator.validate_username(sanitized):
        raise ValueError("Invalid username: contains forbidden characters or too long")
    
    if not sanitized.strip():
        raise ValueError("Username cannot be empty")
    
    return sanitized


def validate_room_id(cls, value: str) -> str:
    if not isinstance(value, str):
        raise ValueError("Room ID must be a string")
    
    sanitized = Sanitizer.sanitize_identifier(value)
    
    if not WhitelistValidator.validate_room_id(sanitized):
        raise ValueError("Invalid room ID: contains forbidden characters or too long")
    
    if not sanitized.strip():
        raise ValueError("Room ID cannot be empty")
    
    return sanitized


def validate_game_action(cls, value: str) -> str:
    if not isinstance(value, str):
        raise ValueError("Action must be a string")
    
    sanitized = Sanitizer.sanitize_action(value)
    
    if not WhitelistValidator.validate_action(sanitized):
        raise ValueError("Invalid action: not in allowed actions list")
    
    return sanitized


def validate_user_id(cls, value) -> int:
    if isinstance(value, str):
        raise ValueError("User ID must be an integer, not a string")
    
    if isinstance(value, float):
        raise ValueError("User ID must be an integer, not a float")
    
    if not isinstance(value, int):
        raise ValueError("User ID must be an integer")
    
    if value < 1 or value > 2147483647:  # Max PostgreSQL integer
        raise ValueError("User ID must be between 1 and 2147483647")
    
    return value


def validate_coordinates(cls, value) -> tuple:
    if not isinstance(value, (tuple, list)) or len(value) != 2:
        raise ValueError("Coordinates must be a tuple/list of 2 elements")
    
    x, y = value
    
    if not isinstance(x, int) or not isinstance(y, int):
        raise ValueError("Coordinates must be integers")
    
    if not (-10000 <= x <= 10000) or not (-10000 <= y <= 10000):
        raise ValueError("Coordinates out of allowed range")
    
    return tuple(value) 


"""
Tests for backpressure handling functionality
"""

import pytest
import asyncio
import json
import time
from unittest.mock import AsyncMock, MagicMock, patch

from aiows import WebSocket, WebSocketServer
from aiows.websocket import BackpressureManager, BackpressureMetrics, SendQueueItem, backpressure_metrics
from aiows.settings import create_settings


class TestBackpressureMetrics:
    """Test backpressure metrics collection"""
    
    def test_initialization(self):
        metrics = BackpressureMetrics()
        
        assert metrics.messages_queued == 0
        assert metrics.messages_sent == 0
        assert metrics.messages_dropped == 0
        assert metrics.queue_overflow_events == 0
        assert metrics.slow_client_disconnections == 0
        assert len(metrics.connection_metrics) == 0
    
    def test_record_message_queued(self):
        metrics = BackpressureMetrics()
        connection_id = "test_conn"
        
        metrics.record_message_queued(connection_id, 5)
        
        assert metrics.messages_queued == 1
        assert connection_id in metrics.connection_metrics
        assert metrics.connection_metrics[connection_id]['messages_queued'] == 1
        assert metrics.connection_metrics[connection_id]['current_queue_size'] == 5
    
    def test_record_message_sent(self):
        metrics = BackpressureMetrics()
        connection_id = "test_conn"
        
        metrics.record_message_sent(connection_id, 100.5, 3)
        
        assert metrics.messages_sent == 1
        assert metrics.total_send_time_ms == 100.5
        assert metrics.connection_metrics[connection_id]['messages_sent'] == 1
        assert metrics.connection_metrics[connection_id]['total_send_time_ms'] == 100.5
    
    def test_record_message_dropped(self):
        metrics = BackpressureMetrics()
        connection_id = "test_conn"
        
        metrics.record_message_dropped(connection_id, "queue_overflow")
        
        assert metrics.messages_dropped == 1
        assert metrics.connection_metrics[connection_id]['messages_dropped'] == 1
        assert metrics.connection_metrics[connection_id]['last_drop_reason'] == "queue_overflow"
    
    def test_get_global_stats(self):
        metrics = BackpressureMetrics()
        
        metrics.record_message_queued("conn1", 1)
        metrics.record_message_sent("conn1", 50.0, 0)
        
        stats = metrics.get_global_stats()
        
        assert 'uptime_seconds' in stats
        assert stats['messages_queued'] == 1
        assert stats['messages_sent'] == 1
        assert stats['average_send_time_ms'] == 50.0
        assert stats['active_connections'] == 1


class TestSendQueueItem:
    """Test send queue item"""
    
    def test_initialization(self):
        data = '{"test": "data"}'
        created_at = time.time()
        
        item = SendQueueItem(data, created_at, "json")
        
        assert item.data == data
        assert item.created_at == created_at
        assert item.item_type == "json"
    
    def test_default_item_type(self):
        item = SendQueueItem("test", time.time())
        assert item.item_type == "data"


class TestBackpressureManager:
    """Test backpressure manager"""
    
    def test_initialization(self):
        manager = BackpressureManager(
            connection_id="test_conn",
            max_queue_size=50,
            overflow_strategy="drop_oldest"
        )
        
        assert manager.connection_id == "test_conn"
        assert manager.max_queue_size == 50
        assert manager.overflow_strategy == "drop_oldest"
        assert manager.queue_size == 0
        assert not manager.is_queue_full
        assert not manager.is_slow_client
    
    def test_queue_properties(self):
        manager = BackpressureManager("test", max_queue_size=10, slow_client_threshold=80)
        
        for i in range(8):
            asyncio.run(manager.enqueue_message(f"msg{i}"))
        
        assert manager.queue_size == 8
        assert manager.queue_utilization_percent == 80.0
        assert manager.is_slow_client
        assert not manager.is_queue_full
    
    @pytest.mark.asyncio
    async def test_enqueue_message_success(self):
        manager = BackpressureManager("test", max_queue_size=5)
        
        success = await manager.enqueue_message("test_data", "data")
        
        assert success is True
        assert manager.queue_size == 1
    
    @pytest.mark.asyncio
    async def test_enqueue_message_overflow_drop_oldest(self):
        manager = BackpressureManager("test", max_queue_size=2, overflow_strategy="drop_oldest")
        
        await manager.enqueue_message("msg1")
        await manager.enqueue_message("msg2")
        
        success = await manager.enqueue_message("msg3")
        
        assert success is True
        assert manager.queue_size == 2
        
        item = await manager.dequeue_message()
        assert "msg2" in item.data
    
    @pytest.mark.asyncio
    async def test_enqueue_message_overflow_drop_newest(self):
        manager = BackpressureManager("test", max_queue_size=2, overflow_strategy="drop_newest")
        
        await manager.enqueue_message("msg1")
        await manager.enqueue_message("msg2")
        
        success = await manager.enqueue_message("msg3")
        
        assert success is False
        assert manager.queue_size == 2
    
    @pytest.mark.asyncio
    async def test_enqueue_message_overflow_reject(self):
        manager = BackpressureManager("test", max_queue_size=2, overflow_strategy="reject")
        
        await manager.enqueue_message("msg1")
        await manager.enqueue_message("msg2")
        
        success = await manager.enqueue_message("msg3")
        
        assert success is False
        assert manager.queue_size == 2
    
    @pytest.mark.asyncio
    async def test_dequeue_message(self):
        manager = BackpressureManager("test")
        
        await manager.enqueue_message("test_data", "data")
        item = await manager.dequeue_message()
        
        assert item is not None
        assert item.data == "test_data"
        assert item.item_type == "data"
        assert manager.queue_size == 0
    
    @pytest.mark.asyncio
    async def test_dequeue_empty_queue(self):
        manager = BackpressureManager("test")
        
        item = await manager.dequeue_message()
        
        assert item is None
    
    def test_slow_client_detection(self):
        manager = BackpressureManager("test", max_queue_size=10, 
                                    slow_client_threshold=50, slow_client_timeout=1.0)
        
        for i in range(4):
            asyncio.run(manager.enqueue_message(f"msg{i}"))
        
        assert not manager.is_slow_client
        assert manager.slow_client_detected_at is None
        
        asyncio.run(manager.enqueue_message("msg_trigger"))
        
        assert manager.is_slow_client
        assert manager.slow_client_detected_at is not None
        
        assert not manager.should_disconnect_slow_client
        
        manager.slow_client_detected_at = time.time() - 2.0
        assert manager.should_disconnect_slow_client
    
    def test_record_successful_send(self):
        manager = BackpressureManager("test", enable_metrics=True)
        
        manager.record_successful_send(150.5)
        
        assert manager.last_successful_send > 0
    
    def test_get_health_stats(self):
        manager = BackpressureManager("test")
        
        stats = manager.get_health_stats()
        
        assert 'connection_id' in stats
        assert 'queue_size' in stats
        assert 'queue_utilization_percent' in stats
        assert 'is_slow_client' in stats
        assert 'time_since_last_send_seconds' in stats


class TestWebSocketBackpressure:
    """Test WebSocket class with backpressure enabled"""
    
    @pytest.fixture
    def mock_websocket(self):
        mock = AsyncMock()
        mock.send = AsyncMock()
        mock.recv = AsyncMock()
        mock.close = AsyncMock()
        return mock
    
    @pytest.fixture
    def backpressure_settings(self):
        return {
            'enabled': True,
            'send_queue_max_size': 10,
            'send_queue_overflow_strategy': 'drop_oldest',
            'slow_client_threshold': 80,
            'slow_client_timeout': 5.0,
            'max_response_time_ms': 1000,
            'enable_send_metrics': True
        }
    
    @pytest.fixture
    def websocket_with_backpressure(self, mock_websocket, backpressure_settings):
        return WebSocket(mock_websocket, backpressure_settings=backpressure_settings)
    
    def test_initialization_with_backpressure(self, websocket_with_backpressure):
        ws = websocket_with_backpressure
        
        assert ws.backpressure_enabled is True
        assert ws._backpressure_manager is not None
    
    def test_initialization_without_backpressure(self, mock_websocket):
        ws = WebSocket(mock_websocket)
        
        assert ws.backpressure_enabled is False
        assert ws._backpressure_manager is None
        assert ws._send_task is None
    
    @pytest.mark.asyncio
    async def test_send_json_with_backpressure(self, websocket_with_backpressure):
        ws = websocket_with_backpressure
        test_data = {"type": "test", "message": "hello"}
        
        await ws.send_json(test_data)
        
        assert ws.send_queue_size == 1
        
        await asyncio.sleep(0.1)
        
        ws._websocket.send.assert_called()
    
    @pytest.mark.asyncio
    async def test_send_with_backpressure(self, websocket_with_backpressure):
        ws = websocket_with_backpressure
        test_data = "test message"
        
        await ws.send(test_data)
        
        assert ws.send_queue_size == 1
        
        await asyncio.sleep(0.1)
        
        ws._websocket.send.assert_called_with(test_data)
    
    @pytest.mark.asyncio
    async def test_send_without_backpressure(self, mock_websocket):
        ws = WebSocket(mock_websocket)
        test_data = "test message"
        
        await ws.send(test_data)
        
        assert ws.send_queue_size == 0
        mock_websocket.send.assert_called_once_with(test_data)
    
    @pytest.mark.asyncio
    async def test_queue_overflow_handling(self, mock_websocket, backpressure_settings):
        backpressure_settings['send_queue_max_size'] = 2
        ws = WebSocket(mock_websocket, backpressure_settings=backpressure_settings)
        
        ws._websocket.send = AsyncMock(side_effect=asyncio.sleep(10))
        
        await ws.send("msg1")
        await ws.send("msg2")
        await ws.send("msg3")
        
        assert ws.send_queue_size == 2
    
    @pytest.mark.asyncio
    async def test_backpressure_stats(self, websocket_with_backpressure):
        ws = websocket_with_backpressure
        
        stats = ws.get_backpressure_stats()
        
        assert stats['backpressure_enabled'] is True
        assert stats['connection_id'] == ws.connection_id
        assert 'queue_size' in stats
        assert 'queue_utilization_percent' in stats
    
    @pytest.mark.asyncio
    async def test_enable_backpressure_runtime(self, mock_websocket, backpressure_settings):
        ws = WebSocket(mock_websocket)
        
        assert ws.backpressure_enabled is False
        
        ws.enable_backpressure(backpressure_settings)
        
        assert ws.backpressure_enabled is True
        assert ws._backpressure_manager is not None
    
    @pytest.mark.asyncio
    async def test_disable_backpressure(self, websocket_with_backpressure):
        ws = websocket_with_backpressure
        
        assert ws.backpressure_enabled is True
        
        ws.disable_backpressure()
        
        assert ws.backpressure_enabled is False
        assert ws._backpressure_manager is None
    
    @pytest.mark.asyncio
    async def test_slow_client_detection_and_disconnect(self, mock_websocket):
        backpressure_settings = {
            'enabled': True,
            'send_queue_max_size': 5,
            'slow_client_threshold': 60,
            'slow_client_timeout': 0.1,
            'enable_send_metrics': True
        }
        
        ws = WebSocket(mock_websocket, backpressure_settings=backpressure_settings)
        
        ws._websocket.send = AsyncMock(side_effect=asyncio.sleep(10))
        
        for i in range(4):
            await ws.send(f"msg{i}")
        
        await asyncio.sleep(0.2)
        
        assert ws.is_closed
    
    @pytest.mark.asyncio
    async def test_cleanup_on_close(self, websocket_with_backpressure):
        ws = websocket_with_backpressure
        
        await ws.send("test_message")
        
        assert ws._send_task is not None
        
        await ws.close()
        
        await asyncio.sleep(0.1)
        
        assert ws._send_task.cancelled() or ws._send_task.done()


class TestWebSocketServerBackpressure:
    """Test WebSocketServer with backpressure integration"""
    
    @pytest.fixture
    def server_with_backpressure(self):
        settings = create_settings("testing")
        settings.backpressure.enabled = True
        settings.backpressure.send_queue_max_size = 5
        
        return WebSocketServer.from_settings(settings)
    
    def test_backpressure_stats_empty(self, server_with_backpressure):
        server = server_with_backpressure
        
        stats = server.get_backpressure_stats()
        
        assert 'global_stats' in stats
        assert 'connection_stats' in stats
        assert len(stats['connection_stats']) == 0
    
    def test_get_slow_connections_empty(self, server_with_backpressure):
        server = server_with_backpressure
        
        slow_connections = server.get_slow_connections()
        
        assert len(slow_connections) == 0


class TestBackpressureIntegration:
    """Integration tests for backpressure functionality"""
    
    @pytest.mark.asyncio
    async def test_end_to_end_backpressure_flow(self):
        """Test complete backpressure flow from server to WebSocket"""
        settings = create_settings("testing")
        settings.backpressure.enabled = True
        settings.backpressure.send_queue_max_size = 3
        settings.backpressure.slow_client_threshold = 66
        
        server = WebSocketServer.from_settings(settings)
        
        mock_websocket = AsyncMock()
        mock_websocket.send = AsyncMock()
        
        backpressure_settings = {
            'enabled': settings.backpressure.enabled,
            'send_queue_max_size': settings.backpressure.send_queue_max_size,
            'send_queue_overflow_strategy': settings.backpressure.send_queue_overflow_strategy,
            'slow_client_threshold': settings.backpressure.slow_client_threshold,
            'slow_client_timeout': settings.backpressure.slow_client_timeout,
            'max_response_time_ms': settings.backpressure.max_response_time_ms,
            'enable_send_metrics': settings.backpressure.enable_send_metrics
        }
        
        ws = WebSocket(mock_websocket, backpressure_settings=backpressure_settings)
        
        await ws.send_json({"test": "message1"})
        await ws.send_json({"test": "message2"})
        
        assert ws.send_queue_size == 2
        assert abs(ws.send_queue_utilization_percent - 66.67) < 0.01
        
        stats = ws.get_backpressure_stats()
        assert stats['is_slow_client'] is True
        
        await ws.close()
    
    @pytest.mark.asyncio 
    async def test_backward_compatibility(self):
        """Test that existing code works without backpressure"""
        mock_websocket = AsyncMock()
        mock_websocket.send = AsyncMock()
        
        ws = WebSocket(mock_websocket)
        
        await ws.send_json({"test": "data"})
        await ws.send("test string")
        
        assert ws.send_queue_size == 0
        assert mock_websocket.send.call_count == 2
        
        assert ws.backpressure_enabled is False
        stats = ws.get_backpressure_stats()
        assert stats['backpressure_enabled'] is False 


"""
Tests for error categories and their behavior across the framework
"""

import asyncio
import socket
import ssl
import pytest
from unittest.mock import Mock, AsyncMock, patch
from aiows.websocket import WebSocket, error_metrics
from aiows.dispatcher import MessageDispatcher, dispatcher_error_metrics
from aiows.router import Router
from aiows.exceptions import (
    ConnectionError, MessageValidationError, MiddlewareError, 
    MessageSizeError, ErrorCategory, ErrorContext, ErrorCategorizer
)
from aiows.middleware.base import BaseMiddleware


class TestFatalErrorHandling:
    """Test FATAL error category behavior"""
    
    @pytest.fixture(autouse=True)
    def reset_metrics(self):
        for attr in dir(error_metrics):
            if attr.endswith('_errors'):
                setattr(error_metrics, attr, 0)
        for attr in dir(dispatcher_error_metrics):
            if attr.endswith('_errors'):
                setattr(dispatcher_error_metrics, attr, 0)
        yield
    
    @pytest.mark.asyncio
    async def test_memory_error_stops_everything(self):
        """Test that MemoryError stops all processing"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        class FatalMiddleware(BaseMiddleware):
            async def on_connect(self, call_next, websocket):
                raise MemoryError("System out of memory")
        
        dispatcher.add_middleware(FatalMiddleware())
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        await dispatcher.dispatch_connect(mock_websocket)
        
        assert dispatcher_error_metrics.fatal_errors == 1
        assert dispatcher_error_metrics.critical_errors == 1
    
    @pytest.mark.asyncio
    async def test_os_error_stops_everything(self):
        """Test that OSError stops all processing"""
        ws = Mock()
        ws.remote_address = ('127.0.0.1', 8080)
        websocket = WebSocket(ws)
        
        ws.send = AsyncMock(side_effect=OSError("System resource unavailable"))
        
        with pytest.raises(ConnectionError):
            await websocket.send("test")
        
        assert websocket.closed
        assert error_metrics.fatal_errors == 1
        assert error_metrics.network_errors == 1
    
    @pytest.mark.asyncio
    async def test_fatal_errors_use_critical_logging(self):
        """Test that FATAL errors use critical log level"""
        memory_error = MemoryError("Critical system failure")
        log_level = ErrorCategorizer.get_log_level(memory_error)
        
        assert log_level == 'critical'
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(memory_error)
        assert should_stop == True


class TestRecoverableErrorHandling:
    """Test RECOVERABLE error category behavior"""
    
    @pytest.fixture(autouse=True)
    def reset_metrics(self):
        for attr in dir(error_metrics):
            if attr.endswith('_errors'):
                setattr(error_metrics, attr, 0)
        for attr in dir(dispatcher_error_metrics):
            if attr.endswith('_errors'):
                setattr(dispatcher_error_metrics, attr, 0)
        yield
    
    @pytest.mark.asyncio
    async def test_timeout_errors_allow_retry(self):
        """Test that timeout errors can be recovered from"""
        ws = Mock()
        ws.remote_address = ('127.0.0.1', 8080)
        websocket = WebSocket(ws)
        
        ws.send = AsyncMock(side_effect=asyncio.TimeoutError("Operation timed out"))
        
        with pytest.raises(ConnectionError):
            await websocket.send("test1")
        
        assert websocket.closed
        assert error_metrics.recoverable_errors == 1
        assert error_metrics.timeout_errors == 1
        
        websocket._reset_connection_state_for_testing()
        
        ws.send = AsyncMock()
        await websocket.send("test2")
        
        assert websocket._error_count == 0
    
    @pytest.mark.asyncio
    async def test_recoverable_errors_use_info_logging(self):
        """Test that RECOVERABLE errors use info log level"""
        timeout_error = asyncio.TimeoutError("Operation timed out")
        log_level = ErrorCategorizer.get_log_level(timeout_error)
        
        assert log_level == 'info'
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(timeout_error)
        assert should_stop == True
    
    @pytest.mark.asyncio
    async def test_graceful_degradation_with_timeout(self):
        """Test graceful degradation when middleware times out"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        call_count = 0
        
        class TimeoutMiddleware(BaseMiddleware):
            async def on_message(self, call_next, websocket, message_data):
                nonlocal call_count
                call_count += 1
                if call_count == 1:
                    raise asyncio.TimeoutError("Middleware timeout")
                await call_next(websocket, message_data)
        
        dispatcher.add_middleware(TimeoutMiddleware())
        
        @router.message()
        async def message_handler(websocket, message):
            websocket.message_processed = True
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        message_data = {"type": "chat", "text": "test1", "user_id": 1}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.recoverable_errors == 1
        
        message_data = {"type": "chat", "text": "test2", "user_id": 1}  
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert hasattr(mock_websocket, 'message_processed')


class TestClientErrorHandling:
    """Test CLIENT_ERROR category behavior"""
    
    @pytest.fixture(autouse=True)
    def reset_metrics(self):
        for attr in dir(error_metrics):
            if attr.endswith('_errors'):
                setattr(error_metrics, attr, 0)
        for attr in dir(dispatcher_error_metrics):
            if attr.endswith('_errors'):
                setattr(dispatcher_error_metrics, attr, 0)
        yield
    
    @pytest.mark.asyncio
    async def test_message_validation_errors_stop_chain_but_preserve_connection(self):
        """Test that validation errors stop middleware chain but don't close connection unnecessarily"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        with pytest.raises(MessageValidationError):
            dispatcher._parse_message_safely({"invalid": "message_structure"})
        
        assert dispatcher_error_metrics.client_errors == 1
        assert dispatcher_error_metrics.parsing_errors == 1
    
    @pytest.mark.asyncio
    async def test_message_size_errors_are_client_errors(self):
        """Test that oversized messages are treated as client errors"""
        ws = Mock()
        ws.remote_address = ('127.0.0.1', 8080)
        websocket = WebSocket(ws, max_message_size=100)
        
        large_message = "x" * 150
        ws.recv = AsyncMock(return_value=large_message)
        
        with pytest.raises(MessageSizeError):
            await websocket.recv()
        
        assert error_metrics.client_errors == 1
        assert error_metrics.size_errors == 1
        assert not websocket.closed
    
    @pytest.mark.asyncio
    async def test_client_errors_use_warning_logging(self):
        """Test that CLIENT errors use warning log level"""
        validation_error = MessageValidationError("Invalid message format")
        log_level = ErrorCategorizer.get_log_level(validation_error)
        
        assert log_level == 'warning'
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(validation_error)
        assert should_stop == True
    
    @pytest.mark.asyncio
    async def test_client_error_response_to_client(self):
        """Test that client errors result in appropriate responses to client"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        class ValidationMiddleware(BaseMiddleware):
            async def on_message(self, call_next, websocket, message_data):
                if not message_data.get("user_id"):
                    raise MessageValidationError("Missing user_id")
                await call_next(websocket, message_data)
        
        dispatcher.add_middleware(ValidationMiddleware())
        
        handler_called = False
        
        @router.message()
        async def message_handler(websocket, message):
            nonlocal handler_called
            handler_called = True
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        message_data = {"type": "chat", "text": "test"}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.client_errors == 1
        assert not handler_called


class TestServerErrorHandling:
    """Test SERVER_ERROR category behavior"""
    
    @pytest.fixture(autouse=True)
    def reset_metrics(self):
        for attr in dir(error_metrics):
            if attr.endswith('_errors'):
                setattr(error_metrics, attr, 0)
        for attr in dir(dispatcher_error_metrics):
            if attr.endswith('_errors'):
                setattr(dispatcher_error_metrics, attr, 0)
        yield
    
    @pytest.mark.asyncio
    async def test_middleware_errors_allow_graceful_degradation(self):
        """Test that middleware errors allow graceful degradation"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        class FailingMiddleware(BaseMiddleware):
            async def on_message(self, call_next, websocket, message_data):
                raise MiddlewareError("Internal middleware failure")
        
        class WorkingMiddleware(BaseMiddleware):
            async def on_message(self, call_next, websocket, message_data):
                websocket.working_middleware_called = True
                await call_next(websocket, message_data)
        
        dispatcher.add_middleware(FailingMiddleware())
        dispatcher.add_middleware(WorkingMiddleware())
        
        @router.message()
        async def message_handler(websocket, message):
            websocket.handler_called = True
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        message_data = {"type": "chat", "text": "test", "user_id": 1}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.server_errors == 1
        assert hasattr(mock_websocket, 'working_middleware_called')
        assert hasattr(mock_websocket, 'handler_called')
    
    @pytest.mark.asyncio
    async def test_connection_errors_stop_processing(self):
        """Test that connection errors stop processing"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        class ConnectionFailingMiddleware(BaseMiddleware):
            async def on_connect(self, call_next, websocket):
                raise ConnectionError("Connection establishment failed")
        
        dispatcher.add_middleware(ConnectionFailingMiddleware())
        
        connect_handler_called = False
        
        @router.connect()
        async def connect_handler(websocket):
            nonlocal connect_handler_called
            connect_handler_called = True
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        await dispatcher.dispatch_connect(mock_websocket)
        
        assert dispatcher_error_metrics.server_errors == 1
        assert not connect_handler_called
    
    @pytest.mark.asyncio
    async def test_server_errors_use_error_logging(self):
        """Test that SERVER errors use error log level"""
        middleware_error = MiddlewareError("Internal server error")
        log_level = ErrorCategorizer.get_log_level(middleware_error)
        
        assert log_level == 'error'
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(middleware_error, "RegularMiddleware")
        assert should_stop == False
    
    @pytest.mark.asyncio
    async def test_critical_middleware_errors_stop_chain(self):
        """Test that auth/security middleware errors stop the chain"""
        auth_error = MiddlewareError("Authentication failed")
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(auth_error, "AuthMiddleware")
        assert should_stop == True
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(auth_error, "SecurityMiddleware")
        assert should_stop == True
        
        should_stop = ErrorCategorizer.should_stop_middleware_chain(auth_error, "LoggingMiddleware")
        assert should_stop == False


class TestErrorContextPreservation:
    """Test that error context is preserved through the call chain"""
    
    @pytest.fixture(autouse=True)
    def reset_metrics(self):
        for attr in dir(error_metrics):
            if attr.endswith('_errors'):
                setattr(error_metrics, attr, 0)
        for attr in dir(dispatcher_error_metrics):
            if attr.endswith('_errors'):
                setattr(dispatcher_error_metrics, attr, 0)
        yield
    
    @pytest.mark.asyncio
    async def test_websocket_error_context_preservation(self):
        """Test that WebSocket errors preserve context through the chain"""
        ws = Mock()
        ws.remote_address = ('127.0.0.1', 8080)
        websocket = WebSocket(ws)
        
        ws.send = AsyncMock(side_effect=socket.error("Network failure"))
        
        with patch.object(websocket, '_log_error_with_context') as mock_log:
            with pytest.raises(ConnectionError):
                await websocket.send("test")
            
            mock_log.assert_called()
            error_arg, context_arg = mock_log.call_args[0]
            
            assert isinstance(error_arg, socket.error)
            assert hasattr(context_arg, 'operation')
            assert hasattr(context_arg, 'component')
            assert hasattr(context_arg, 'error_id')
            assert context_arg.operation == 'send'
            assert context_arg.component == 'websocket'
    
    @pytest.mark.asyncio
    async def test_dispatcher_error_context_preservation(self):
        """Test that Dispatcher errors preserve context through the chain"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        class ContextTestMiddleware(BaseMiddleware):
            async def on_message(self, call_next, websocket, message_data):
                raise ValueError("Test error with context")
        
        dispatcher.add_middleware(ContextTestMiddleware())
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        with patch.object(dispatcher, '_log_error_with_context') as mock_log:
            message_data = {"type": "chat", "text": "test", "user_id": 1}
            await dispatcher.dispatch_message(mock_websocket, message_data)
            
            mock_log.assert_called()
            error_arg, context_arg = mock_log.call_args[0]
            
            assert isinstance(error_arg, ValueError)
            assert hasattr(context_arg, 'operation')
            assert hasattr(context_arg, 'component')
            assert hasattr(context_arg, 'error_id')
            assert 'middleware_message' in context_arg.operation
            assert context_arg.component == 'dispatcher'
            assert 'middleware_name' in context_arg.additional_context
    
    @pytest.mark.asyncio
    async def test_error_id_consistency(self):
        """Test that error IDs are consistent and unique"""
        context1 = ErrorContext(operation="test1", component="test")
        context2 = ErrorContext(operation="test2", component="test")
        
        assert context1.error_id != context2.error_id
        
        assert context1.error_id == context1.error_id
        
        custom_context = ErrorContext(operation="test", component="test", error_id="CUSTOM123")
        assert custom_context.error_id == "CUSTOM123"


class TestIntegrationErrorHandling:
    """Integration tests for error handling across all components"""
    
    @pytest.fixture(autouse=True)
    def reset_metrics(self):
        for attr in dir(error_metrics):
            if attr.endswith('_errors'):
                setattr(error_metrics, attr, 0)
        for attr in dir(dispatcher_error_metrics):
            if attr.endswith('_errors'):
                setattr(dispatcher_error_metrics, attr, 0)
        yield
    
    @pytest.mark.asyncio
    async def test_full_error_handling_workflow(self):
        """Test complete error handling workflow from WebSocket to Dispatcher"""
        router = Router()
        dispatcher = MessageDispatcher(router)
        
        class MultiErrorMiddleware(BaseMiddleware):
            def __init__(self):
                self.call_count = 0
            
            async def on_message(self, call_next, websocket, message_data):
                self.call_count += 1
                message_type = message_data.get('error_type')
                
                if message_type == 'client':
                    raise MessageValidationError("Client sent invalid data")
                elif message_type == 'server':
                    raise MiddlewareError("Server middleware error")
                elif message_type == 'recoverable':
                    if self.call_count == 1:
                        raise asyncio.TimeoutError("Temporary timeout")
                elif message_type == 'fatal':
                    raise MemoryError("System out of memory")
                
                await call_next(websocket, message_data)
        
        middleware = MultiErrorMiddleware()
        dispatcher.add_middleware(middleware)
        
        handler_processed = False
        
        @router.message()
        async def message_handler(websocket, message):
            nonlocal handler_processed
            handler_processed = True
        
        mock_websocket = Mock()
        mock_websocket.context = {}
        mock_websocket.remote_address = ('127.0.0.1', 8080)
        mock_websocket.closed = False
        mock_websocket.close = AsyncMock()
        
        message_data = {"type": "chat", "text": "test", "user_id": 1, "error_type": "client"}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.client_errors == 1
        assert not handler_processed
        
        handler_processed = False
        
        message_data = {"type": "chat", "text": "test", "user_id": 1, "error_type": "server"}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.server_errors == 1
        assert handler_processed
        
        handler_processed = False
        
        middleware.call_count = 0
        
        message_data = {"type": "chat", "text": "test", "user_id": 1, "error_type": "recoverable"}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.recoverable_errors == 1
        assert not handler_processed
        
        handler_processed = False
        message_data = {"type": "chat", "text": "test", "user_id": 1, "error_type": "recoverable"}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert handler_processed
        
        message_data = {"type": "chat", "text": "test", "user_id": 1, "error_type": "fatal"}
        await dispatcher.dispatch_message(mock_websocket, message_data)
        
        assert dispatcher_error_metrics.fatal_errors == 1
    
    @pytest.mark.asyncio
    async def test_metrics_consistency_across_components(self):
        """Test that metrics are consistent across WebSocket and Dispatcher components"""
        validation_error = MessageValidationError("Test validation error")
        connection_error = ConnectionError("Test connection error")
        memory_error = MemoryError("Test memory error")
        timeout_error = asyncio.TimeoutError("Test timeout error")
        
        assert ErrorCategorizer.categorize_exception(validation_error) == ErrorCategory.CLIENT_ERROR
        assert ErrorCategorizer.categorize_exception(connection_error) == ErrorCategory.SERVER_ERROR
        assert ErrorCategorizer.categorize_exception(memory_error) == ErrorCategory.FATAL
        assert ErrorCategorizer.categorize_exception(timeout_error) == ErrorCategory.RECOVERABLE
        
        assert ErrorCategorizer.get_log_level(validation_error) == 'warning'
        assert ErrorCategorizer.get_log_level(connection_error) == 'error'
        assert ErrorCategorizer.get_log_level(memory_error) == 'critical'
        assert ErrorCategorizer.get_log_level(timeout_error) == 'info'
        
        assert ErrorCategorizer.should_stop_middleware_chain(validation_error) == True
        assert ErrorCategorizer.should_stop_middleware_chain(connection_error) == True
        assert ErrorCategorizer.should_stop_middleware_chain(memory_error) == True
        assert ErrorCategorizer.should_stop_middleware_chain(timeout_error) == True
        
        regular_error = MiddlewareError("Regular error")
        assert ErrorCategorizer.should_stop_middleware_chain(regular_error, "LoggingMiddleware") == False 


import pytest
import asyncio
import json
import logging
import threading
import socket
import time
from io import StringIO

try:
    import websockets
except ImportError:
    pytest.skip("websockets library not available", allow_module_level=True)

from aiows import (
    WebSocketServer, 
    Router, 
    WebSocket, 
    BaseMessage,
    AuthMiddleware,
    LoggingMiddleware,
    RateLimitingMiddleware
)
from aiows.middleware.auth import generate_auth_token

def get_free_port():
    """Get a free port for testing"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(('', 0))
        return s.getsockname()[1]

class RuntimeTestServer:
    """Helper for setting up test servers with middleware"""
    
    def __init__(self):
        self.server = None
        self.port = None
        self.server_thread = None
        self.router = None
        
    def setup_basic_router(self):
        router = Router()
        
        @router.connect()
        async def handle_connect(websocket: WebSocket):
            user_id = websocket.context.get('user_id', 'anonymous')
            authenticated = websocket.context.get('authenticated', False)
            
            await websocket.send_json({
                "type": "connected",
                "user_id": user_id,
                "authenticated": authenticated
            })
        
        @router.message("test")
        async def handle_test(websocket: WebSocket, message: BaseMessage):
            user_id = websocket.context.get('user_id', 'anonymous')
            rate_limit = websocket.context.get('rate_limit', {})
            
            await websocket.send_json({
                "type": "test_response",
                "user_id": user_id,
                "remaining_messages": rate_limit.get('remaining_messages', 'unknown'),
                "echo": message.dict()
            })
            
        @router.disconnect()
        async def handle_disconnect(websocket: WebSocket, reason: str):
            pass
            
        self.router = router
        return router
    
    def start_server_with_middleware(self, middleware_list):
        self.port = get_free_port()
        
        def run_server():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            
            self.server = WebSocketServer()
            
            for middleware in middleware_list:
                self.server.add_middleware(middleware)
            
            router = self.setup_basic_router()
            self.server.include_router(router)
            
            try:
                self.server.run(host="localhost", port=self.port)
            except Exception as e:
                print(f"Server error: {e}")
        
        self.server_thread = threading.Thread(target=run_server, daemon=True)
        self.server_thread.start()
        time.sleep(0.5)
        
        return f"ws://localhost:{self.port}"
    
    def stop(self):
        if self.server_thread:
            self.server_thread.join(timeout=1)

@pytest.fixture
def test_runtime_server():
    server = RuntimeTestServer()
    yield server
    server.stop()

class TestAuthMiddlewareRuntime:
    
    @pytest.mark.asyncio
    async def test_valid_token_authentication(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=3)
        uri = test_runtime_server.start_server_with_middleware([auth_middleware])
        
        jwt_token = generate_auth_token(
            user_id="user123",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        valid_uri = f"{uri}?token={jwt_token}"
        
        async with websockets.connect(valid_uri) as websocket:
            response = await websocket.recv()
            data = json.loads(response)
            
            assert data["type"] == "connected"
            assert data["user_id"] == "user123"
            assert data["authenticated"] == True
    
    @pytest.mark.asyncio
    async def test_valid_token_in_headers(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=3)
        uri = test_runtime_server.start_server_with_middleware([auth_middleware])
        
        jwt_token = generate_auth_token(
            user_id="user456",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        headers = {"Authorization": f"Bearer {jwt_token}"}
        
        async with websockets.connect(uri, additional_headers=headers) as websocket:
            response = await websocket.recv()
            data = json.loads(response)
            
            assert data["type"] == "connected"
            assert data["user_id"] == "user456"
            assert data["authenticated"] == True
    
    @pytest.mark.asyncio
    async def test_invalid_token_closes_connection(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=1)
        uri = test_runtime_server.start_server_with_middleware([auth_middleware])
        
        invalid_uri = f"{uri}?token=invalid_token"
        
        with pytest.raises(websockets.exceptions.ConnectionClosedError) as exc_info:
            async with websockets.connect(invalid_uri) as websocket:
                await websocket.recv()
        
        assert exc_info.value.code == 4401
    
    @pytest.mark.asyncio
    async def test_no_token_closes_connection(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=1)
        uri = test_runtime_server.start_server_with_middleware([auth_middleware])
        
        with pytest.raises(websockets.exceptions.ConnectionClosed) as exc_info:
            async with websockets.connect(uri) as websocket:
                await websocket.recv()
        
        assert exc_info.value.code in [1000, 4401]
    
    @pytest.mark.asyncio
    async def test_authenticated_messaging(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=3)
        uri = test_runtime_server.start_server_with_middleware([auth_middleware])
        
        jwt_token = generate_auth_token(
            user_id="testuser",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        valid_uri = f"{uri}?token={jwt_token}"
        
        async with websockets.connect(valid_uri) as websocket:
            await websocket.recv()
            await websocket.send(json.dumps({"type": "test", "data": "hello"}))
            response = await websocket.recv()
            data = json.loads(response)
            
            assert data["type"] == "test_response"
            assert data["user_id"] == "testuser"

class TestLoggingMiddlewareRuntime:
    
    def setup_log_capture(self):
        log_stream = StringIO()
        handler = logging.StreamHandler(log_stream)
        handler.setLevel(logging.INFO)
        
        logger = logging.getLogger("aiows.test")
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
        
        return log_stream, logger
    
    @pytest.mark.asyncio
    async def test_connection_logging(self, test_runtime_server):
        log_stream, logger = self.setup_log_capture()
        
        logging_middleware = LoggingMiddleware("aiows.test")
        uri = test_runtime_server.start_server_with_middleware([logging_middleware])
        
        async with websockets.connect(uri) as websocket:
            await websocket.recv()
            await asyncio.sleep(0.1)
        
        logs = log_stream.getvalue()
        assert "WebSocket connection established" in logs
    
    @pytest.mark.asyncio
    async def test_message_logging(self, test_runtime_server):
        log_stream, logger = self.setup_log_capture()
        
        logging_middleware = LoggingMiddleware("aiows.test")
        uri = test_runtime_server.start_server_with_middleware([logging_middleware])
        
        async with websockets.connect(uri) as websocket:
            await websocket.recv()
            await websocket.send(json.dumps({"type": "test", "data": "test_message"}))
            await websocket.recv()
            await asyncio.sleep(0.1)
        
        logs = log_stream.getvalue()
        assert "Message received" in logs
        assert "Message processed" in logs

class TestRateLimitingMiddlewareRuntime:
    
    @pytest.mark.asyncio
    async def test_rate_limit_enforcement(self, test_runtime_server):
        rate_middleware = RateLimitingMiddleware(max_messages_per_minute=2)
        uri = test_runtime_server.start_server_with_middleware([rate_middleware])
        
        async with websockets.connect(uri) as websocket:
            await websocket.recv()
            await websocket.send(json.dumps({"type": "test", "data": "msg1"}))
            response1 = json.loads(await websocket.recv())
            assert response1["remaining_messages"] == 1
            
            await websocket.send(json.dumps({"type": "test", "data": "msg2"}))
            response2 = json.loads(await websocket.recv())
            assert response2["remaining_messages"] == 0
            
            await websocket.send(json.dumps({"type": "test", "data": "msg3"}))
            
            with pytest.raises(websockets.exceptions.ConnectionClosedError) as exc_info:
                await websocket.recv()
            
            assert exc_info.value.code == 4429
    
    @pytest.mark.asyncio
    async def test_rate_limit_with_user_identification(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=3)
        rate_middleware = RateLimitingMiddleware(max_messages_per_minute=1)
        
        uri = test_runtime_server.start_server_with_middleware([auth_middleware, rate_middleware])
        
        jwt_token = generate_auth_token(
            user_id="user1",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        valid_uri = f"{uri}?token={jwt_token}"
        async with websockets.connect(valid_uri) as ws1:
            await ws1.recv()
            await ws1.send(json.dumps({"type": "test", "data": "user1_msg"}))
            await ws1.recv()
            
        jwt_token = generate_auth_token(
            user_id="user2",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        user2_uri = f"{uri}?token={jwt_token}"
        
        async with websockets.connect(user2_uri) as ws2:
            await ws2.recv()
            await ws2.send(json.dumps({"type": "test", "data": "user2_msg"}))
            response = json.loads(await ws2.recv())
            assert response["user_id"] == "user2"

class TestMiddlewareChainRuntime:
    
    @pytest.mark.asyncio
    async def test_middleware_execution_order(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=3)
        logging_middleware = LoggingMiddleware("aiows.chain")
        rate_middleware = RateLimitingMiddleware(max_messages_per_minute=10)
        
        middleware_chain = [auth_middleware, logging_middleware, rate_middleware]
        uri = test_runtime_server.start_server_with_middleware(middleware_chain)
        
        jwt_token = generate_auth_token(
            user_id="userX",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        valid_uri = f"{uri}?token={jwt_token}"
        
        async with websockets.connect(valid_uri) as websocket:
            response = await websocket.recv()
            data = json.loads(response)
            
            assert data["authenticated"] == True
            assert data["user_id"] == "userX"
            
            await websocket.send(json.dumps({"type": "test", "data": "chain_test"}))
            response = json.loads(await websocket.recv())
            
            assert response["user_id"] == "userX"
            assert "remaining_messages" in response
    
    @pytest.mark.asyncio
    async def test_middleware_error_stops_chain(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=1)
        rate_middleware = RateLimitingMiddleware(max_messages_per_minute=10)
        
        middleware_chain = [auth_middleware, rate_middleware]
        uri = test_runtime_server.start_server_with_middleware(middleware_chain)
        
        with pytest.raises(websockets.exceptions.ConnectionClosed) as exc_info:
            async with websockets.connect(uri) as websocket:
                await websocket.recv()
        
        assert exc_info.value.code in [1000, 4401]
    
    @pytest.mark.asyncio 
    async def test_context_preservation_across_middleware(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=3)
        rate_middleware = RateLimitingMiddleware(max_messages_per_minute=5)
        
        middleware_chain = [auth_middleware, rate_middleware]
        uri = test_runtime_server.start_server_with_middleware(middleware_chain)
        
        jwt_token = generate_auth_token(
            user_id="contextuser",
            secret_key=secret_key,
            ttl_seconds=300
        )
        
        valid_uri = f"{uri}?token={jwt_token}"
        
        async with websockets.connect(valid_uri) as websocket:
            await websocket.recv()
            await websocket.send(json.dumps({"type": "test", "data": "context_test"}))
            response = json.loads(await websocket.recv())
            
            assert response["user_id"] == "contextuser"
            assert response["remaining_messages"] == 4

class TestMiddlewarePerformance:
    
    @pytest.mark.asyncio
    async def test_rapid_connection_attempts(self, test_runtime_server):
        secret_key = "test_secret_key_32_characters_long_12345"
        auth_middleware = AuthMiddleware(secret_key, auth_timeout=1)
        uri = test_runtime_server.start_server_with_middleware([auth_middleware])
        
        tasks = []
        for i in range(5):
            async def try_connect():
                try:
                    async with websockets.connect(uri) as ws:
                        await ws.recv()
                except websockets.exceptions.ConnectionClosed:
                    return "closed"
                return "connected"
            
            tasks.append(try_connect())
        
        results = await asyncio.gather(*tasks, return_exceptions=True)
        
        for result in results:
            assert result == "closed"
    
    @pytest.mark.asyncio
    async def test_concurrent_rate_limiting(self, test_runtime_server):
        rate_middleware = RateLimitingMiddleware(max_messages_per_minute=3)
        uri = test_runtime_server.start_server_with_middleware([rate_middleware])
        
        async with websockets.connect(uri) as ws1, websockets.connect(uri) as ws2:
            await ws1.recv()
            await ws2.recv()
            
            await ws1.send(json.dumps({"type": "test", "data": "ws1_msg1"}))
            await ws2.send(json.dumps({"type": "test", "data": "ws2_msg1"}))
            
            response1 = json.loads(await ws1.recv())
            response2 = json.loads(await ws2.recv())
            
            assert response1["remaining_messages"] == 2
            assert response2["remaining_messages"] == 2

if __name__ == "__main__":
    pytest.main([__file__, "-v", "--tb=short"]) 


"""
Factory functions for creating aiows components from configuration
"""

from typing import List, Optional
from .server import WebSocketServer
from .settings import AiowsSettings, create_settings
from .middleware.base import BaseMiddleware
from .middleware.auth import AuthMiddleware
from .middleware.rate_limit import RateLimitingMiddleware
from .middleware.connection_limiter import ConnectionLimiterMiddleware
from .middleware.logging import LoggingMiddleware


def create_server_from_settings(settings: Optional[AiowsSettings] = None, 
                               include_middleware: bool = True) -> WebSocketServer:
    """Create WebSocketServer with automatic middleware configuration
    
    Args:
        settings: AiowsSettings instance (creates default if None)
        include_middleware: Whether to automatically add configured middleware
        
    Returns:
        Configured WebSocketServer instance
    """
    if settings is None:
        settings = create_settings()
    
    server = WebSocketServer.from_settings(settings)
    
    if include_middleware:
        middleware_list = create_middleware_from_settings(settings)
        for middleware in middleware_list:
            server.add_middleware(middleware)
    
    return server


def create_middleware_from_settings(settings: AiowsSettings) -> List[BaseMiddleware]:
    """Create list of middleware instances from settings
    
    Args:
        settings: AiowsSettings instance with configuration
        
    Returns:
        List of configured middleware instances in correct order
    """
    middleware_list = []
    
    # Order matters: Logging -> ConnectionLimiter -> Auth -> RateLimit
    if settings.logging.enabled:
        middleware_list.append(LoggingMiddleware.from_config(settings.logging))
    
    if settings.connection_limiter.enabled:
        middleware_list.append(ConnectionLimiterMiddleware.from_config(settings.connection_limiter))
    
    if settings.auth.enabled:
        middleware_list.append(AuthMiddleware.from_config(settings.auth))
    
    if settings.rate_limit.enabled:
        middleware_list.append(RateLimitingMiddleware.from_config(settings.rate_limit))
    
    return middleware_list


def create_production_server(profile: str = "production") -> WebSocketServer:
    """Create production-ready server with all security features enabled
    
    IMPORTANT: Production servers require SSL configuration!
    Either provide ssl_context when creating the server or set SSL certificate paths
    in environment variables (AIOWS_SSL_CERT_FILE, AIOWS_SSL_KEY_FILE).
    
    Args:
        profile: Configuration profile to use (default: "production")
        
    Returns:
        Production-configured WebSocketServer with all middleware
        
    Raises:
        ValueError: If SSL context is required but not provided
    """
    settings = AiowsSettings(profile=profile)
    
    if profile == "production":
        settings.auth.enabled = True
        settings.connection_limiter.enabled = True
        settings.rate_limit.enabled = True
        settings.logging.enabled = True
        settings.logging.sanitize_data = True
        settings.logging.use_json_format = True
        # SSL requirement is already correctly set to True in production profile
        # settings.server.require_ssl_in_production = True  # Already set by profile
    
    return create_server_from_settings(settings, include_middleware=True)


def create_development_server(profile: str = "development") -> WebSocketServer:
    """Create development server with relaxed security for testing
    
    Args:
        profile: Configuration profile to use (default: "development")
        
    Returns:
        Development-configured WebSocketServer
    """
    settings = AiowsSettings(profile=profile)
    
    if profile == "development":
        settings.rate_limit.max_messages_per_minute = 120
        settings.connection_limiter.max_connections_per_ip = 20
        settings.logging.log_level = "DEBUG"
        settings.logging.include_message_content = True
        settings.logging.sanitize_data = False
    
    return create_server_from_settings(settings, include_middleware=True)


def create_testing_server(profile: str = "testing") -> WebSocketServer:
    """Create testing server with minimal overhead and fast timeouts
    
    Args:
        profile: Configuration profile to use (default: "testing")
        
    Returns:
        Testing-configured WebSocketServer
    """
    settings = AiowsSettings(profile=profile)
    
    if profile == "testing":
        settings.server.shutdown_timeout = 1.0
        settings.server.cleanup_interval = 0.1
        settings.rate_limit.max_messages_per_minute = 1000
        settings.connection_limiter.max_connections_per_ip = 100
        settings.logging.log_level = "ERROR"
        settings.logging.enable_rate_limiting = False
        settings.auth.auth_timeout = 1
    
    return create_server_from_settings(settings, include_middleware=True)


def create_custom_server(host: str = None,
                        port: int = None,
                        profile: str = "development",
                        enable_auth: bool = False,
                        enable_logging: bool = True,
                        enable_rate_limit: bool = True,
                        enable_connection_limiter: bool = True) -> WebSocketServer:
    """Create server with custom configuration
    
    Args:
        host: Server host (uses profile default if None)
        port: Server port (uses profile default if None)
        profile: Base configuration profile
        enable_auth: Whether to enable authentication
        enable_logging: Whether to enable logging middleware
        enable_rate_limit: Whether to enable rate limiting
        enable_connection_limiter: Whether to enable connection limiting
        
    Returns:
        Custom-configured WebSocketServer
    """
    settings = AiowsSettings(profile=profile)
    
    if host is not None:
        settings.server.host = host
    if port is not None:
        settings.server.port = port
    
    settings.auth.enabled = enable_auth
    settings.logging.enabled = enable_logging
    settings.rate_limit.enabled = enable_rate_limit
    settings.connection_limiter.enabled = enable_connection_limiter
    
    return create_server_from_settings(settings, include_middleware=True)


def create_auth_middleware(settings: AiowsSettings) -> AuthMiddleware:
    return AuthMiddleware.from_config(settings.auth)


def create_rate_limit_middleware(settings: AiowsSettings) -> RateLimitingMiddleware:
    return RateLimitingMiddleware.from_config(settings.rate_limit)


def create_connection_limiter_middleware(settings: AiowsSettings) -> ConnectionLimiterMiddleware:
    return ConnectionLimiterMiddleware.from_config(settings.connection_limiter)


def create_logging_middleware(settings: AiowsSettings) -> LoggingMiddleware:
    return LoggingMiddleware.from_config(settings.logging)


def auto_configure_for_environment() -> WebSocketServer:
    """Automatically configure server based on environment variables
    
    Reads AIOWS_PROFILE environment variable or detects environment:
    - If AIOWS_PROFILE=production -> production server
    - If AIOWS_PROFILE=testing -> testing server
    - Default -> development server
    
    Returns:
        Auto-configured WebSocketServer
    """
    import os
    
    profile = os.getenv('AIOWS_PROFILE', 'development')
    
    if profile == "production":
        return create_production_server()
    elif profile == "testing":
        return create_testing_server()
    else:
        return create_development_server()


def quick_start_server(host: str = "localhost", 
                      port: int = 8000,
                      with_auth: bool = False) -> WebSocketServer:
    """Quick start server for rapid prototyping
    
    Args:
        host: Server host
        port: Server port  
        with_auth: Whether to enable authentication
        
    Returns:
        Ready-to-use WebSocketServer for development
    """
    return create_custom_server(
        host=host,
        port=port,
        profile="development",
        enable_auth=with_auth,
        enable_logging=True,
        enable_rate_limit=True,
        enable_connection_limiter=True
    ) 